__all__ = ["DistributionChart"]

from typing import *

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
from matplotlib.animation import FuncAnimation
from numpy import ndarray

from . import BaseChart
from . import normal_color
from ..metrics import loss_numpy
from ..models import SSUResult, ArtificialSample
from ..statistics import to_microns, mode


class DistributionChart(BaseChart):
    def __init__(self, parent=None, size=(3, 2.5)):
        super().__init__(parent=parent, figsize=size)
        self.setWindowTitle(self.tr("Distribution Chart"))
        self._axes: plt.Axes = self._figure.subplots()
        self.scale_menu = QtWidgets.QMenu(self.tr("Scale"))
        self.menu.insertMenu(self.edit_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions: List[QtGui.QAction] = []
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name)
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)
        self.show_mode_lines_action = QtGui.QAction(self.tr("Show Mode Lines"))
        self.show_mode_lines_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.show_mode_lines_action)
        self.show_mode_lines_action.setCheckable(True)
        self.show_mode_lines_action.setChecked(False)
        self.show_legend_action = QtGui.QAction(self.tr("Show Legend"))
        self.show_legend_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.show_legend_action)
        self.show_legend_action.setCheckable(True)
        self.show_legend_action.setChecked(False)
        self.animated_action = QtGui.QAction(self.tr("Animated"))
        self.animated_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.animated_action)
        self.animated_action.setCheckable(True)
        self.animated_action.setChecked(False)
        self.interval_menu = QtWidgets.QMenu(self.tr("Animation Interval"))
        self.menu.insertMenu(self.edit_figure_action, self.interval_menu)
        self.interval_group = QtGui.QActionGroup(self.interval_menu)
        self.interval_group.setExclusive(True)
        self.interval_actions: List[QtGui.QAction] = []
        for interval, name in self.supported_intervals:
            interval_action = self.interval_group.addAction(name)
            interval_action.setCheckable(True)
            interval_action.triggered.connect(self.update_chart)
            self.interval_menu.addAction(interval_action)
            self.interval_actions.append(interval_action)
        self.interval_actions[3].setChecked(True)
        self.repeat_animation_action = QtGui.QAction(self.tr("Repeat Animation"))
        self.repeat_animation_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.repeat_animation_action)
        self.repeat_animation_action.setCheckable(True)
        self.repeat_animation_action.setChecked(False)
        self._last_result: Union[ArtificialSample, SSUResult, None] = None

    @property
    def supported_scales(self) -> Sequence[Tuple[str, str]]:
        scales = (("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear")))
        return scales

    @property
    def supported_intervals(self) -> Sequence[Tuple[int, str]]:
        intervals = ((5, self.tr("5 Milliseconds")),
                     (10, self.tr("10 Milliseconds")),
                     (20, self.tr("20 Milliseconds")),
                     (30, self.tr("30 Milliseconds")),
                     (60, self.tr("60 Milliseconds")))
        return intervals

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def show_mode(self) -> bool:
        return self.show_mode_lines_action.isChecked()

    @property
    def show_legend(self) -> bool:
        return self.show_legend_action.isChecked()

    @property
    def animated(self) -> bool:
        return self.animated_action.isChecked()

    @property
    def animation_interval(self) -> int:
        for i, interval_action in enumerate(self.interval_actions):
            if interval_action.isChecked():
                interval, name = self.supported_intervals[i]
                return interval

    @property
    def repeat_animation(self) -> bool:
        return self.repeat_animation_action.isChecked()

    @property
    def transfer(self) -> Callable[[Union[float, ndarray]], Union[float, ndarray]]:
        if self.scale == "log-linear":
            return lambda classes_phi: to_microns(classes_phi)
        elif self.scale == "log":
            return lambda classes_phi: np.log(to_microns(classes_phi))
        elif self.scale == "phi":
            return lambda classes_phi: classes_phi
        elif self.scale == "linear":
            return lambda classes_phi: to_microns(classes_phi)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size (microns)"
        elif self.scale == "log":
            return "Ln(grain size in microns)"
        elif self.scale == "phi":
            return "Grain size (phi)"
        elif self.scale == "linear":
            return "Grain size (microns)"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def show_menu(self, pos: QtCore.QPoint):
        self.edit_figure_action.setEnabled(self._animation is None and self._last_result is not None)
        self.save_figure_action.setEnabled(self._animation is None and self._last_result is not None)
        self.menu.popup(QtGui.QCursor.pos())

    def show_chart(self, result: Union[ArtificialSample, SSUResult]):
        self._last_result = result
        self._axes.clear()
        if self._animation is not None:
            self._animation._stop()
            self._animation = None
        x = self.transfer(result.classes_phi)
        self._axes.set_title(result.name)
        self._axes.set_xlabel(self.xlabel)
        self._axes.set_ylabel(self.ylabel)
        self._axes.plot(x, result.sample.distribution, c="#ffffff00", marker=".", ms=3, mfc=normal_color(),
                        mec=normal_color(), label="Observation")
        self._axes.set_xlim(x[0], x[-1])
        self._axes.set_ylim(0.0, round(np.max(result.sample.distribution) * 1.2, 2))
        lmse_loss = loss_numpy("lmse")(result.distribution, result.sample.distribution, None)
        self._axes.plot(x, result.distribution, c=normal_color(), label=f"Prediction (LMSE={lmse_loss:.2f})")
        for i, component in enumerate(result):
            mode_micron = mode(result.classes, result.classes_phi, component.distribution, is_geometric=True)
            self._axes.plot(x, component.distribution * component.proportion, c=plt.get_cmap()(i),
                            label=f"C{i + 1} ({mode_micron:.2f} μm, {component.proportion:.2%}))")
        if self.xlog:
            self._axes.set_xscale("log")
        if self.show_mode:
            modes = [self.transfer(mode(result.classes, result.classes_phi, component.distribution,
                                        is_geometric=False)) for component in result]
            colors = [plt.get_cmap()(i) for i in range(len(result))]
            self._axes.vlines(modes, 0.0, 1.0, colors=colors)
        if self.show_legend:
            self._axes.legend(loc="upper left")
        self._figure.tight_layout()
        self._canvas.draw()

    def show_animation(self, result: SSUResult):
        assert isinstance(result, SSUResult)
        assert result.n_iterations > 1
        self._last_result = result
        self._axes.clear()
        if self._animation is not None:
            self._animation._stop()
            self._animation = None
        x = self.transfer(result.classes_phi)
        if self.xlog:
            self._axes.set_xscale("log")
        self._axes.set_title(result.name)
        self._axes.set_xlabel(self.xlabel)
        self._axes.set_ylabel(self.ylabel)
        observation_line = self._axes.plot(x, result.sample.distribution, c="#ffffff00", marker=".", ms=3,
                                           mfc=normal_color(), mec=normal_color(), label="Observation")[0]
        self._axes.set_xlim(x[0], x[-1])
        self._axes.set_ylim(0.0, round(np.max(result.sample.distribution) * 1.2, 2))

        prediction_line: Optional[plt.Line2D] = None
        component_lines: List[plt.Line2D] = []
        mode_lines: Optional[plt.Artist] = None
        legend: Optional[plt.Artist] = None

        def init():
            nonlocal prediction_line
            nonlocal component_lines
            nonlocal mode_lines
            nonlocal legend
            if prediction_line is None:
                lmse_loss = loss_numpy("lmse")(result.distribution, result.sample.distribution, None)
                prediction_line = self._axes.plot(x, result.distribution, c=normal_color(),
                                                  label=f"Prediction (LMSE={lmse_loss:.2f})")[0]
                for i, component in enumerate(result):
                    mode_micron = mode(result.classes, result.classes_phi, component.distribution, is_geometric=True)
                    line = self._axes.plot(x, component.distribution * component.proportion, c=plt.get_cmap()(i),
                                           label=f"C{i + 1} ({mode_micron:.2f} μm, {component.proportion:.2%}))")[0]
                    component_lines.append(line)
                if self.show_mode:
                    modes = [self.transfer(mode(result.classes, result.classes_phi, component.distribution,
                                                is_geometric=False)) for component in result]
                    colors = [plt.get_cmap()(i) for i in range(len(result))]
                    mode_lines = self._axes.vlines(modes, 0.0, 1.0, colors=colors)
                if self.show_legend:
                    legend = self._axes.legend(loc="upper left")
            artists = [prediction_line, *component_lines]
            if mode_lines is not None:
                artists.append(mode_lines)
            if legend is not None:
                artists.append(legend)
            return artists

        def animate(current: SSUResult):
            nonlocal prediction_line
            nonlocal component_lines
            nonlocal mode_lines
            nonlocal legend
            prediction_line.set_ydata(current.distribution)
            for i, (line, component) in enumerate(zip(component_lines, current)):
                mode_micron = mode(current.classes, current.classes_phi, component.distribution, is_geometric=True)
                line.set_ydata(component.distribution * component.proportion)
                line.set_label(f"C{i + 1} ({mode_micron:.2f} μm, {component.proportion:.2%}))")
            artists = [prediction_line, *component_lines]
            if self.show_mode:
                mode_lines.remove()
                modes = [self.transfer(mode(current.classes, current.classes_phi, component.distribution,
                                            is_geometric=False)) for component in current]
                colors = [plt.get_cmap()(i) for i in range(len(current))]
                mode_lines = self._axes.vlines(modes, 0.0, 1.0, colors=colors)
                artists.append(mode_lines)
            if self.show_legend:
                lmse_loss = loss_numpy("lmse")(current.distribution, current.sample.distribution, None)
                handles = [observation_line, prediction_line, *component_lines]
                labels = ["Observation", f"Prediction (LMSE={lmse_loss:.2f})"]
                for i, component in enumerate(current):
                    mode_micron = mode(current.classes, current.classes_phi, component.distribution, is_geometric=True)
                    label = f"C{i + 1} ({mode_micron:.2f} μm, {component.proportion:.2%})"
                    labels.append(label)
                legend = self._axes.legend(handles=handles, labels=labels, loc="upper left")
                artists.append(legend)
            return artists

        self._animation = FuncAnimation(self._figure, animate, frames=result.history, init_func=init,
                                        interval=self.animation_interval, blit=True, repeat=self.repeat_animation,
                                        repeat_delay=5.0, save_count=result.n_iterations)

    def show_result(self, result: Union[ArtificialSample, SSUResult]):
        if self.animated and isinstance(result, SSUResult) and result.n_iterations > 1:
            self.show_animation(result)
        else:
            self.show_chart(result)

    def update_chart(self):
        self._figure.clear()
        self._axes = self._figure.subplots()
        if self._last_result is not None:
            self.show_result(self._last_result)

    def retranslate(self):
        self.setWindowTitle(self.tr("Distribution Chart"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))
        self.scale_menu.setTitle(self.tr("Scale"))
        for action, (key, name) in zip(self.scale_actions, self.supported_scales):
            action.setText(name)
        self.show_mode_lines_action.setText(self.tr("Show Mode Lines"))
        self.show_legend_action.setText(self.tr("Show Legend"))
        self.animated_action.setText(self.tr("Animated"))
        self.interval_menu.setTitle(self.tr("Animation Interval"))
        for action, (interval, name) in zip(self.interval_actions, self.supported_intervals):
            action.setText(name)
        self.repeat_animation_action.setText(self.tr("Repeat Animation"))



__all__ = ["LossSeriesChart"]

from typing import *

from . import BaseChart


class LossSeriesChart(BaseChart):
    def __init__(self, parent=None, figsize=(3, 2.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Loss Series"))
        self._axes = self._figure.subplots()
        self._last_result = None

    def show_loss_series(self, series: Sequence[float], ylabel: str, title: str = ""):
        self._axes.clear()
        self._axes.plot(series, label="series")
        self._axes.set_xlabel("Iteration index")
        self._axes.set_ylabel(ylabel)
        self._axes.set_title(title)
        self._figure.tight_layout()
        self._canvas.draw()
        self._last_result = series, ylabel, title

    def update_chart(self):
        if self._last_result is not None:
            self._figure.clear()
            self._axes = self._figure.subplots()
            self.show_loss_series(*self._last_result)

    def retranslate(self):
        self.setWindowTitle(self.tr("Loss Series"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))



__all__ = ["Normal", "SkewNormal", "Weibull", "GeneralWeibull",
           "get_distribution", "get_sorted_indexes", "sort_parameters"]

import typing

import numpy as np
from scipy.special import softmax
from scipy.stats import norm, skewnorm, weibull_min

from .models import DistributionType

_INFINITESIMAL = 1e-8


def relu(x):
    return np.maximum(x, _INFINITESIMAL)


class Normal:
    NAME = "Normal"
    N_PARAMETERS = 2
    PARAMETER_NAMES = ("Location", "Scale")
    PARAMETER_BOUNDS = ((None, None),
                        (_INFINITESIMAL, None))

    @staticmethod
    def interpret(parameters: np.ndarray, classes: np.ndarray, interval: float):
        n_samples, n_components, n_classes = classes.shape
        assert parameters.ndim == 3
        assert parameters.shape == (n_samples, Normal.N_PARAMETERS + 1, n_components)
        locations = np.expand_dims(parameters[:, 0, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(parameters[:, 1, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(parameters[:, 2, :], axis=1), 1)
        components = norm.pdf(classes, loc=locations, scale=scales) * interval
        m, v, s, k = norm.stats(loc=locations[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, (m, np.sqrt(v), s, k)

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((Normal.N_PARAMETERS + 1, n_components))
        defaults[0] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[1] = np.random.random((n_components,)) * 0.1 + 2.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults


class SkewNormal:
    NAME = "Skew Normal"
    N_PARAMETERS = 3
    PARAMETER_NAMES = ("Shape", "Location", "Scale")
    PARAMETER_BOUNDS = ((None, None),
                        (None, None),
                        (_INFINITESIMAL, None))

    @staticmethod
    def interpret(parameters: np.ndarray, classes: np.ndarray, interval: float):
        n_samples, n_components, n_classes = classes.shape
        assert parameters.ndim == 3
        assert parameters.shape == (n_samples, SkewNormal.N_PARAMETERS + 1, n_components)
        shapes = np.expand_dims(parameters[:, 0, :], 2).repeat(n_classes, 2)
        locations = np.expand_dims(parameters[:, 1, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(parameters[:, 2, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(parameters[:, 3, :], axis=1), 1)
        components = skewnorm.pdf(classes, shapes, loc=locations, scale=scales) * interval
        m, v, s, k = skewnorm.stats(shapes[:, :, 0], loc=locations[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, (m, np.sqrt(v), s, k)

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((SkewNormal.N_PARAMETERS + 1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1
        defaults[1] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        defaults[3] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults


class Weibull:
    NAME = "Weibull"
    N_PARAMETERS = 2
    PARAMETER_NAMES = ("Shape", "Scale")
    PARAMETER_BOUNDS = ((_INFINITESIMAL, None),
                        (_INFINITESIMAL, None))

    @staticmethod
    def interpret(parameters: np.ndarray, classes: np.ndarray, interval: float):
        n_samples, n_components, n_classes = classes.shape
        assert parameters.ndim == 3
        assert parameters.shape == (n_samples, Weibull.N_PARAMETERS + 1, n_components)
        shapes = np.expand_dims(relu(parameters[:, 0, :]), 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(parameters[:, 1, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(parameters[:, 2, :], axis=1), 1)
        components = weibull_min.pdf(classes, shapes, scale=scales) * interval
        m, v, s, k = weibull_min.stats(shapes[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, (m, np.sqrt(v), s, k)

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((Weibull.N_PARAMETERS + 1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1 + 3.60234942
        defaults[1] = np.random.random((n_components,)) * 0.1 + 3.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults


class GeneralWeibull:
    NAME = "General Weibull"
    N_PARAMETERS = 3
    PARAMETER_NAMES = ("Shape", "Location", "Scale")
    PARAMETER_BOUNDS = ((_INFINITESIMAL, None),
                        (None, None),
                        (_INFINITESIMAL, None))

    @staticmethod
    def interpret(parameters: np.ndarray, classes: np.ndarray, interval: float):
        n_samples, n_components, n_classes = classes.shape
        assert parameters.ndim == 3
        assert parameters.shape == (n_samples, GeneralWeibull.N_PARAMETERS + 1, n_components)
        shapes = np.expand_dims(relu(parameters[:, 0, :]), 2).repeat(n_classes, 2)
        locations = np.expand_dims(parameters[:, 1, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(parameters[:, 2, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(parameters[:, 3, :], axis=1), 1)
        components = weibull_min.pdf(classes, shapes, loc=locations, scale=scales) * interval
        m, v, s, k = weibull_min.stats(shapes[:, :, 0], loc=locations[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, (m, np.sqrt(v), s, k)

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((GeneralWeibull.N_PARAMETERS + 1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1 + 3.60234942
        defaults[1] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 3.0
        defaults[3] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults


def get_distribution(distribution_type: DistributionType):
    if distribution_type == DistributionType.Normal:
        return Normal
    elif distribution_type == DistributionType.SkewNormal:
        return SkewNormal
    elif distribution_type == DistributionType.Weibull:
        return Weibull
    elif distribution_type == DistributionType.GeneralWeibull:
        return GeneralWeibull
    else:
        raise NotImplementedError(distribution_type)


def get_sorted_indexes(
        distribution_type: DistributionType,
        parameters: np.ndarray,
        classes: np.ndarray,
        interval: float) -> typing.Tuple[int]:
    distribution_class = get_distribution(distribution_type)
    proportions, components, (m, std, s, k) = distribution_class.interpret(parameters, classes, interval)
    mean_values = [(i, mean) for i, mean in enumerate(np.median(m, axis=0))]
    # sort them by mean size
    mean_values.sort(key=lambda x: x[1], reverse=True)
    sorted_indexes = tuple([i for i, _ in mean_values])
    return sorted_indexes


def sort_parameters(
        distribution_type: DistributionType,
        parameters: np.ndarray,
        classes: np.ndarray,
        interval: float) -> np.ndarray:
    sorted_indexes = get_sorted_indexes(distribution_type, parameters, classes, interval)
    sorted_parameters = np.zeros_like(parameters)
    for i, j in enumerate(sorted_indexes):
        sorted_parameters[:, :, i] = parameters[:, :, j]
    return sorted_parameters



__all__ = ["EMMAResult"]

import copy
from typing import *

import numpy as np
from numpy import ndarray

from ..models import KernelType, Dataset, ArtificialDataset
from ..metrics import loss_numpy


class EMMAResult:
    """The class to represent the result of EMMA algorithm."""
    def __init__(self, dataset: [ArtificialDataset, Dataset], kernel_type: KernelType, n_members: int,
                 proportions: ndarray, end_members: ndarray, time_spent: Union[int, float], x0: ndarray = None,
                 settings: Dict[str, Any] = None, loss_series: Dict[str, ndarray] = None):
        # do some validations
        assert isinstance(dataset, (ArtificialDataset, Dataset))
        assert isinstance(kernel_type, KernelType)
        assert isinstance(n_members, int)
        assert isinstance(proportions, ndarray)
        n_iterations = len(proportions)
        assert proportions.ndim == 3
        assert proportions.shape == (n_iterations, len(dataset), n_members)
        assert isinstance(end_members, ndarray)
        assert end_members.ndim == 3
        assert end_members.shape == (n_iterations, n_members, len(dataset.classes))
        assert isinstance(time_spent, (int, float))
        if x0 is not None:
            assert isinstance(x0, ndarray)
            assert x0.ndim == 2
            assert x0.shape[1] == n_members
        if settings is not None:
            assert isinstance(settings, dict)
            if loss_series is not None:
                assert settings["loss"] in loss_series
                if settings["need_history"]:
                    assert len(loss_series[settings["loss"]]) == proportions.shape[0]

        self._dataset = dataset
        self._kernel_type = kernel_type
        self._n_members = n_members
        self._x0 = x0
        self._proportions = proportions
        self._end_members = end_members
        self._i = -1
        self._time_spent = time_spent
        self._settings = settings
        self._loss_series = loss_series
        modes = [(i, dataset.classes[np.unravel_index(np.argmax(end_members[-1, i]), end_members[-1, i].shape)])
                 for i in range(n_members)]
        modes.sort(key=lambda x: x[1])
        self._sorted_indexes = tuple([i for i, _ in modes])
        self._sort()

    @property
    def dataset(self) -> Union[ArtificialDataset, Dataset]:
        return self._dataset

    @property
    def n_samples(self) -> int:
        return len(self._dataset)

    @property
    def n_members(self) -> int:
        return self._n_members

    @property
    def n_classes(self) -> int:
        return len(self._dataset.classes_phi)

    @property
    def kernel_type(self) -> KernelType:
        return self._kernel_type

    @property
    def proportions(self) -> ndarray:
        return self._proportions[self._i]

    @property
    def end_members(self) -> ndarray:
        return self._end_members[self._i]

    @property
    def distributions(self) -> ndarray:
        return self.proportions @ self.end_members

    @property
    def time_spent(self) -> Union[int, float]:
        return self._time_spent

    @property
    def x0(self) -> Optional[ndarray]:
        return self._x0

    @property
    def n_iterations(self) -> int:
        return self._proportions.shape[0]

    @property
    def history(self):
        for i in range(self._proportions.shape[0]):
            copy_result = copy.copy(self)
            copy_result._i = i
            yield copy_result

    @property
    def settings(self) -> Optional[Dict[str, Any]]:
        if self._settings is None:
            return None
        else:
            return self._settings.copy()

    def _sort(self):
        proportions = np.zeros_like(self._proportions)
        end_members = np.zeros_like(self._end_members)
        for i, j in enumerate(self._sorted_indexes):
            proportions[:, :, i] = self._proportions[:, :, j]
            end_members[:, i, :] = self._end_members[:, j, :]
        self._proportions = proportions
        self._end_members = end_members

    def loss(self, name: str) -> float:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = self.proportions @ self.end_members
        return loss_func(prediction, observation, None)

    def loss_series(self, name: str) -> ndarray:
        if name in self._loss_series:
            return self._loss_series[name]
        loss_func = loss_numpy(name)
        try:
            observation = np.expand_dims(self._dataset.distributions, 0).repeat(self.n_iterations, axis=0)
            prediction = self._proportions @ self._end_members
            series = loss_func(prediction, observation, (1, 2))
            self._loss_series[name] = series
            return series
        except MemoryError:
            series = []
            for result in self.history:
                series.append(result.loss(name))
            series = np.array(series)
            self._loss_series[name] = series
            return series

    def class_wise_losses(self, name: str) -> ndarray:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = self.proportions @ self.end_members
        losses = loss_func(prediction, observation, 0)
        return losses

    def sample_wise_losses(self, name: str) -> ndarray:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = self.proportions @ self.end_members
        losses = loss_func(prediction, observation, 1)
        return losses



__all__ = ["EMMAResultChart"]

from typing import *

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
from matplotlib.animation import FuncAnimation
from numpy import ndarray

from . import BaseChart
from . import normal_color
from ..models import EMMAResult
from ..statistics import to_microns
from ..utils import get_image_by_proportions


class EMMAResultChart(BaseChart):
    N_DISPLAY_SAMPLES = 200

    def __init__(self, parent=None, figsize=(4.4, 4.4)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("EMMA Chart"))
        # self.axes = self.figure.subplots()
        self.scale_menu = QtWidgets.QMenu(self.tr("Scale"))
        self.menu.insertMenu(self.edit_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions: List[QtGui.QAction] = []
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name)
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)
        self.loss_menu = QtWidgets.QMenu(self.tr("Loss Function"))
        self.menu.insertMenu(self.edit_figure_action, self.loss_menu)
        self.loss_group = QtGui.QActionGroup(self.loss_menu)
        self.loss_group.setExclusive(True)
        self.loss_actions: List[QtGui.QAction] = []
        for key, name in self.supported_losses:
            loss_action = self.loss_group.addAction(name)
            loss_action.setCheckable(True)
            loss_action.triggered.connect(self.update_chart)
            self.loss_menu.addAction(loss_action)
            self.loss_actions.append(loss_action)
        self.loss_actions[8].setChecked(True)
        self.animated_action = QtGui.QAction(self.tr("Animated"))
        self.animated_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.animated_action)
        self.animated_action.setCheckable(True)
        self.animated_action.setChecked(False)
        self.interval_menu = QtWidgets.QMenu(self.tr("Animation Interval"))
        self.menu.insertMenu(self.edit_figure_action, self.interval_menu)
        self.interval_group = QtGui.QActionGroup(self.interval_menu)
        self.interval_group.setExclusive(True)
        self.interval_actions: List[QtGui.QAction] = []
        for interval, name in self.supported_intervals:
            interval_action = self.interval_group.addAction(name)
            interval_action.setCheckable(True)
            interval_action.triggered.connect(self.update_chart)
            self.interval_menu.addAction(interval_action)
            self.interval_actions.append(interval_action)
        self.interval_actions[3].setChecked(True)
        self.repeat_action = QtGui.QAction(self.tr("Repeat Animation"))
        self.repeat_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.edit_figure_action, self.repeat_action)
        self.repeat_action.setCheckable(True)
        self.repeat_action.setChecked(False)
        self._last_result: Optional[EMMAResult] = None

    @property
    def supported_scales(self) -> Sequence[Tuple[str, str]]:
        scales = (("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear")))
        return scales

    @property
    def supported_losses(self) -> Sequence[Tuple[str, str]]:
        losses = (("1-norm", self.tr("1 Norm")),
                  ("2-norm", self.tr("2 Norm")),
                  ("3-norm", self.tr("3 Norm")),
                  ("4-norm", self.tr("4 Norm")),
                  ("mae", self.tr("MAE")),
                  ("mse", self.tr("MSE")),
                  ("rmse", self.tr("RMSE")),
                  ("rmlse", self.tr("RMLSE")),
                  ("lmse", self.tr("LMSE")),
                  ("cosine", self.tr("Cosine")),
                  ("angular", self.tr("Angular")))
        return losses

    @property
    def supported_intervals(self) -> Sequence[Tuple[int, str]]:
        intervals = ((5, self.tr("5 Milliseconds")),
                     (10, self.tr("10 Milliseconds")),
                     (20, self.tr("20 Milliseconds")),
                     (30, self.tr("30 Milliseconds")),
                     (60, self.tr("60 Milliseconds")))
        return intervals

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def loss(self) -> Tuple[str, str]:
        for i, distance_action in enumerate(self.loss_actions):
            if distance_action.isChecked():
                key, name = self.supported_losses[i]
                return (key, name)

    @property
    def animated(self) -> bool:
        return self.animated_action.isChecked()

    @property
    def animation_interval(self) -> int:
        for i, interval_action in enumerate(self.interval_actions):
            if interval_action.isChecked():
                interval, name = self.supported_intervals[i]
                return interval

    @property
    def repeat_animation(self) -> bool:
        return self.repeat_action.isChecked()

    @property
    def transfer(self) -> Callable[[Union[int, float, ndarray]], Union[int, float, ndarray]]:
        if self.scale == "log-linear":
            return lambda classes_phi: to_microns(classes_phi)
        elif self.scale == "log":
            return lambda classes_phi: np.log(to_microns(classes_phi))
        elif self.scale == "phi":
            return lambda classes_phi: classes_phi
        elif self.scale == "linear":
            return lambda classes_phi: to_microns(classes_phi)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size (microns)"
        elif self.scale == "log":
            return "Ln(grain size in microns)"
        elif self.scale == "phi":
            return "Grain size (phi)"
        elif self.scale == "linear":
            return "Grain size (microns)"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def show_menu(self, pos: QtCore.QPoint):
        self.edit_figure_action.setEnabled(self._last_result is not None and not self.animated)
        self.save_figure_action.setEnabled(self._last_result is not None and not self.animated)
        self.menu.popup(QtGui.QCursor.pos())

    def show_chart(self, result: EMMAResult):
        self._last_result = result
        self._figure.clear()
        if self._animation is not None:
            self._animation._stop()
            self._animation = None
        classes = self.transfer(result.dataset.classes_phi)
        loss_key, loss_name = self.loss
        loss_series = result.loss_series(loss_key)
        iteration_indexes = np.linspace(1, len(loss_series), len(loss_series))
        interval = max(1, result.n_samples // self.N_DISPLAY_SAMPLES)
        sample_axes = self._figure.add_subplot(2, 2, 1)
        for sample in result.dataset[::interval]:
            sample_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            sample_axes.set_xscale("log")
        sample_axes.set_xlim(classes[0], classes[-1])
        sample_axes.set_ylim(0.0, round(np.max(result.dataset.distributions) * 1.2, 2))
        sample_axes.set_xlabel(self.xlabel)
        sample_axes.set_ylabel(self.ylabel)
        sample_axes.set_title("GSDs")
        loss_axes = self._figure.add_subplot(2, 2, 2)
        loss_axes.plot(iteration_indexes, loss_series, c=normal_color())
        loss_axes.set_xlim(0, len(loss_series))
        loss_axes.set_xlabel("Iteration")
        loss_axes.set_ylabel(loss_name)
        loss_axes.set_title("Loss variation")
        end_member_axes = self._figure.add_subplot(2, 2, 3)
        if self.xlog:
            end_member_axes.set_xscale("log")
        for i in range(result.n_members):
            end_member_axes.plot(classes, result.end_members[i], c=plt.get_cmap()(i), label=f"EM{i + 1}", zorder=10 + i)
        end_member_axes.set_xlim(classes[0], classes[-1])
        end_member_axes.set_ylim(0.0, round(np.max(result.end_members) * 1.2, 2))
        end_member_axes.set_xlabel(self.xlabel)
        end_member_axes.set_ylabel(self.ylabel)
        end_member_axes.set_title("End members")
        proportion_axes = self._figure.add_subplot(2, 2, 4)
        image = get_image_by_proportions(result.proportions, resolution=100)
        proportion_axes.imshow(image, plt.get_cmap(), aspect="auto", vmin=0, vmax=9,
                               extent=(0.0, result.n_samples, 100, 0.0), interpolation="none")
        proportion_axes.set_xlim(0, result.n_samples)
        proportion_axes.set_ylim(0, 100)
        proportion_axes.set_yticks([0, 20, 40, 60, 80, 100], ["0.0", "0.2", "0.4", "0.6", "0.8", "1.0"])
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")
        self._figure.tight_layout()
        self._canvas.draw()

    def show_animation(self, result: EMMAResult):
        assert result.n_iterations > 1
        self._last_result = result
        self._figure.clear()
        if self._animation is not None:
            self._animation._stop()
            self._animation = None
        classes = self.transfer(result.dataset.classes_phi)
        loss_key, loss_name = self.loss
        loss_series = result.loss_series(loss_key)
        iteration_indexes = np.linspace(1, len(loss_series), len(loss_series))
        loss_series = result.loss_series(self.loss[0])
        min_distance, max_distance = np.min(loss_series), np.max(loss_series)
        interval = max(1, result.n_samples // self.N_DISPLAY_SAMPLES)
        sample_axes = self._figure.add_subplot(2, 2, 1)
        for sample in result.dataset[::interval]:
            sample_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            sample_axes.set_xscale("log")
        sample_axes.set_xlim(classes[0], classes[-1])
        sample_axes.set_ylim(0.0, round(np.max(result.dataset.distributions) * 1.2, 2))
        sample_axes.set_xlabel(self.xlabel)
        sample_axes.set_ylabel(self.ylabel)
        sample_axes.set_title("GSDs")
        loss_axes = self._figure.add_subplot(2, 2, 2)
        loss_axes.plot(iteration_indexes, loss_series, c=normal_color())
        loss_axes.set_xlim(0, len(loss_series))
        loss_axes.set_xlabel("Iteration")
        loss_axes.set_ylabel(loss_name)
        loss_axes.set_title("Loss variation")
        end_member_axes = self._figure.add_subplot(2, 2, 3)
        if self.xlog:
            end_member_axes.set_xscale("log")
        end_member_axes.set_xlim(classes[0], classes[-1])
        end_member_axes.set_ylim(0, round(np.max(result.end_members) * 1.2, 2))
        end_member_axes.set_xlabel(self.xlabel)
        end_member_axes.set_ylabel(self.ylabel)
        end_member_axes.set_title("End members")
        proportion_axes = self._figure.add_subplot(2, 2, 4)
        proportion_axes.set_xlim(0, result.n_samples)
        proportion_axes.set_ylim(0, 100)
        proportion_axes.set_yticks([0, 20, 40, 60, 80, 100], ["0.0", "0.2", "0.4", "0.6", "0.8", "1.0"])
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")

        iteration_line: Optional[plt.Line2D] = None
        end_member_curves: List[plt.Line2D] = []
        proportion_image: Optional[plt.Artist] = None

        def init():
            nonlocal iteration_line
            nonlocal end_member_curves
            nonlocal proportion_image
            if iteration_line is None:
                iteration_line = loss_axes.plot([1, 1], [min_distance, max_distance], c=normal_color())[0]
                for i in range(result.n_members):
                    curve = end_member_axes.plot(classes, result.end_members[i],
                                                 c=plt.get_cmap()(i), label=f"EM{i + 1}")[0]
                    end_member_curves.append(curve)
                image = get_image_by_proportions(result.proportions, resolution=100)
                proportion_image = proportion_axes.imshow(
                    image, plt.get_cmap(), aspect="auto", vmin=0, vmax=9,
                    extent=(0.0, result.n_samples, 100, 0.0), interpolation="none")
            return iteration_line, proportion_image, *end_member_curves

        def animate(args: Tuple[int, EMMAResult]):
            nonlocal iteration_line
            nonlocal end_member_curves
            nonlocal proportion_image
            iteration, current = args
            iteration_line.set_xdata((iteration, iteration))
            for i in range(current.n_members):
                end_member_curves[i].set_ydata(current.end_members[i])
            image = get_image_by_proportions(current.proportions, resolution=100)
            proportion_image.set_data(image)
            return iteration_line, proportion_image, *end_member_curves

        self._animation = FuncAnimation(self._figure, animate, init_func=init, frames=enumerate(result.history),
                                        interval=self.animation_interval, blit=True, repeat=self.repeat_animation,
                                        repeat_delay=5.0, save_count=result.n_iterations)

    def show_result(self, result: EMMAResult):
        if self.animated and result.n_iterations > 1:
            self.show_animation(result)
        else:
            self.show_chart(result)

    def update_chart(self):
        if self._last_result is not None:
            self.show_result(self._last_result)

    def retranslate(self):
        self.setWindowTitle(self.tr("EMMA Chart"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))
        self.scale_menu.setTitle(self.tr("Scale"))
        for action, (key, name) in zip(self.scale_actions, self.supported_scales):
            action.setText(name)
        self.loss_menu.setTitle(self.tr("Loss Function"))
        for action, (key, name) in zip(self.loss_actions, self.supported_losses):
            action.setText(name)
        self.animated_action.setText(self.tr("Animated"))
        self.interval_menu.setTitle(self.tr("Animation Interval"))
        for action, (interval, name) in zip(self.interval_actions, self.supported_intervals):
            action.setText(name)
        self.repeat_action.setText(self.tr("Repeat Animation"))



from setuptools import find_packages, setup

setup(
    name="QGrain",
    version="0.5.3.4",
    description="An easy-to-use software for the comprehensive analysis of grain size distributions",
    platforms="all",
    author="Yuming Liu",
    author_email="liuyuming@ieecas.cn",
    url="https://github.com/yuriok/QGrain",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "psutil",
        "numpy",
        "scipy",
        "scikit-learn",
        "grpcio",
        "protobuf",
        "xlrd",
        "openpyxl",
        "PySide6>=6.2.0",
        "matplotlib>=3.5.0",
        "SciencePlots<=1.0.8",
        "qt-material"],
    extras_require={"server": ["torch"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: English"

    ],
    entry_points={
        "console_scripts": ["qgrain=QGrain:main"]
    }
)



__all__ = ["validate_classes", "validate_distributions", "Sample", "Dataset"]

from typing import *

import numpy as np
from numpy import ndarray

from ..statistics import interval_phi


def _incremental(classes: Sequence[Union[int, float]]) -> Tuple[bool, Optional[int]]:
    """
    Check if the series of grain size classes is incremental.

    :param classes: The grain size classes.
    :returns:
        is_incremental: If the series of classes is incremental.
        error_index: If it is incremental, return `None`, else return the index of first invalid value.
    """
    classes = tuple(classes)
    for i, (left, right) in enumerate(zip(classes[:-1], classes[1:])):
        if left >= right:
            return False, i
    else:
        return True, None


def _error_text(array: ndarray, index: int):
    w = 3
    l, r = max(index - w, 0), min(index + w, len(array))
    l_header = [] if l == 0 else ["..."]
    r_header = [] if r == len(array) else ["..."]
    indicator = [""] * len(array)
    indicator[index] = "↑"
    indicator[min(index + 1, len(array) - 1)] = "↑"
    error_text = "\n".join(
        ["\t".join(["Index   "] + l_header + [f"{i:<10}" for i in range(len(array))[l:r]] + r_header),
         "\t".join(["Value   "] + l_header + [f"{f'{v:.4f}':<10}" for v in array[l:r]] + r_header),
         "\t".join(["Position"] + l_header + [f"{s:<10}" for s in indicator[l:r]] + r_header)])
    return error_text


def validate_classes(classes: Sequence[float]) -> Tuple[bool, Union[ndarray, str]]:
    """
    Check if the series of grain size classes is valid.

    :param classes: The grain size classes.
    :returns:
        is_valid: If the series of classes is valid.
        array_or_msg: If it is valid, return a new numpy array, else return the error message.
    """
    if classes is None:
        return False, "The passed classes can not be `None`."
    try:
        array: ndarray = np.array(classes, dtype=np.float64)
    except ValueError as e:
        return False, "Can not convert the classes to a numerical array, " \
                      f"it may contains invalid values (e.g. text). {e}"
    if array.ndim != 1:
        return False, "The passed classes should be one-dimensional."
    if len(array) == 0:
        return False, "The passed classes can not be empty."
    indices = np.arange(len(array))
    nan_indices: ndarray = indices[np.isnan(array)]
    if len(nan_indices) > 0:
        return False, (f"There is at least one NaN value in the series of grain size classes. "
                       f"Check the index(es): {', '.join(nan_indices.astype(str))}.")
    incremental, index = _incremental(array)
    if not incremental:
        error_text = _error_text(array, index)
        return False, f"The series of grain size classes is not incremental.\n{error_text}"
    classes_phi = -np.log2(array / 1000)
    mean_interval = interval_phi(classes_phi)
    intervals = classes_phi[:-1] - classes_phi[1:]
    absolute_errors = np.abs(intervals - mean_interval)
    index = np.argmax(absolute_errors)
    if absolute_errors[index] > 0.05:
        error_text = _error_text(array, index)
        return False, (f"The grain size classes are not evenly spaced on a log scale. "
                       f"The max absolute error of intervals is {absolute_errors[index]} phi.\n{error_text}")
    return True, array


def validate_distributions(distributions: Sequence[Sequence[float]]) -> Tuple[bool, Union[ndarray, str]]:
    """
    Check if the data of grain size distributions is valid.

    :param distributions: The grain size distributions.
    :returns:
        is_valid: If the data is valid.
        array_or_msg: If it is valid, return a new numpy array, else return the error message.
    """
    if distributions is None:
        return False, "The passed distributions can not be `None`."
    try:
        array: ndarray = np.array(distributions, dtype=np.float64)
    except ValueError as e:
        return False, "Can not convert the distributions to a numerical array, " \
                      f"it may contains invalid values (e.g. text). {e}"
    if array.ndim != 2:
        return False, "The passed distributions should be two-dimensional."
    n_samples, n_classes = array.shape
    if n_samples == 0 or n_classes == 0:
        return False, "The passed distribution can not be empty."
    cols, rows = np.meshgrid(np.arange(n_classes), np.arange(n_samples))
    nan_keys = np.isnan(array)
    cells = [f"({row}, {col})" for row, col in zip(rows[nan_keys], cols[nan_keys])]
    if len(cells) > 0:
        return False, (f"There is at least one NaN value in the passed distributions. "
                       f"See the cell(s): {', '.join(cells)}")
    rows = np.arange(n_samples)
    summed = np.sum(array, axis=1)
    valid = (np.greater(summed, 0.95) & np.less(summed, 1.05)) | (np.greater(summed, 95) & np.less(summed, 105))
    invalid_rows = rows[np.logical_not(valid)]
    if len(invalid_rows) > 0:
        return False, (f"There is at least one distribution which has the sum not equal to 1 or 100. "
                       f"Check the following row(s): {', '.join(invalid_rows.astype(str))}.")
    array = array / np.sum(array, axis=1, keepdims=True)
    return True, array


class Sample:
    """The class to represent one sample of the grain size dataset."""
    __slots__ = ("_name", "_classes", "_classes_phi", "_distribution")

    def __init__(self, name: str,
                 classes: ndarray,
                 classes_phi: ndarray,
                 distribution: ndarray):
        """
        Construct an instance of the `Sample` class.

        **If not necessary, do not manually create the sample, because it will not validate the passed parameters.**

        :param name: The name of this sample.
        :param classes: The grain size classes in microns.
        :param classes_phi: The grain size classes in phi values.
        :param distribution: The frequency distribution of grain size classes.
            Note, the sum of frequencies should be equal to 1.
        :return: An instance of the `Sample` class.
        """
        self._name = name
        self._classes = classes
        self._classes_phi = classes_phi
        self._distribution = distribution

    def __repr__(self):
        return f"Sample({self._name})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self._classes_phi)

    @property
    def distribution(self) -> ndarray:
        return self._distribution


class Dataset:
    """
    The class to represent the grain size dataset.

    * Get the sample at index i, `sample = dataset[i]`.
    * Iterate all samples, `for sample in dataset`.
    * Iterate partial samples, `for sample in dataset[:10]`.
    * Get the number of samples, `len(dataset)`.
    """
    def __init__(self, name: str, sample_names: Sequence[str],
                 classes: Sequence[Union[int, float]],
                 distributions: Sequence[Sequence[Union[int, float]]]):
        """
        Construct an instance of the `Dataset` class.

        :param name: The name of this dataset.
        :param sample_names: The names of samples in this dataset.
        :param classes: The grain size classes in microns.
        :param distributions: The grain size distributions of all samples.
            Note, the sum of frequencies of each sample should be equal to 1.
        :return: An instance of the `Dataset` class.
        :raises TypeError: If the name of dataset or any sample is not a string.
        :raises ValueError: If the name of dataset or any sample is empty.
            If any value in the grain size classes or distributions is invalid.
            You can call the functions, `validate_classes` and `validate_distributions`, to check them before.
        """
        if not isinstance(name, str):
            raise TypeError("The name of dataset must be a string.")
        if len(name) == 0:
            raise ValueError("The name of dataset can not be empty.")
        if len(sample_names) != len(distributions):
            raise ValueError("The lengths of sample names and distributions are not equal.")
        for i, sample_name in enumerate(sample_names):
            if not isinstance(sample_name, str):
                raise TypeError(f"The name of sample must be a string. This error raised at the index: {i}.")
            if len(sample_name) == 0:
                raise ValueError(f"The name of sample can not be empty. This error raised at the index: {i}.")
        valid, array_or_msg = validate_classes(classes)
        if not valid:
            raise ValueError(array_or_msg)
        self._classes = array_or_msg
        valid, array_or_msg = validate_distributions(distributions)
        if not valid:
            raise ValueError(array_or_msg)
        self._distributions = array_or_msg
        self._name = name
        self._classes_phi = -np.log2(self._classes / 1000)
        self._sample_names = list(sample_names)

    def __repr__(self) -> str:
        return f"Dataset({self._name})"

    def __len__(self) -> int:
        return len(self._sample_names)

    def __iter__(self):
        for i in range(len(self._sample_names)):
            yield self._get_sample(i)

    def __getitem__(self, item) -> Union[Sample, Sequence[Sample]]:
        if isinstance(item, int):
            return self._get_sample(item)
        elif isinstance(item, slice):
            return [self._get_sample(index) for index in np.arange(len(self._sample_names))[item]]
        else:
            raise TypeError(f"Sample indices must be integers or slices, not {type(item)}.")

    @property
    def name(self) -> str:
        return self._name

    @property
    def sample_names(self) -> List[str]:
        return self._sample_names.copy()

    @property
    def n_classes(self) -> int:
        return len(self._classes)

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self.classes_phi)

    @property
    def distributions(self) -> ndarray:
        return self._distributions

    def _get_sample(self, index: int) -> Sample:
        name = self._sample_names[index]
        distribution = self._distributions[index]
        sample = Sample(name, self._classes, self._classes_phi, distribution)
        return sample



__all__ = ["loss_torch", "NonparametricKernel", "NormalKernel", "SkewNormalKernel", "WeibullKernel",
           "GeneralWeibullKernel", "ProportionModule", "get_kernel"]

from typing import *

import numpy as np
import torch
from torch import Tensor

from .models import KernelType

_INFINITESIMAL = 1e-8


# P-Norm
def p_norm_torch(values: Tensor, targets: Tensor, p=2) -> Tensor:
    return torch.sum(torch.abs(values - targets) ** p) ** (1 / p)


# Mean Absolute Error
def mae_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.mean(torch.abs(values - targets))


# Mean Squared Error
def mse_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.mean(torch.square(values - targets))


# Root Mean Squared Error
def rmse_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.sqrt(torch.mean(torch.square(values - targets)))


# Root Mean Squared Logarithmic Error
def rmlse_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.sqrt(torch.mean(torch.square(torch.log(values + 1) - torch.log(targets + 1))))


# Logarithmic Mean Squared Error
def lmse_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.log(torch.mean(torch.square(values - targets)))


# Cosine
def cosine_torch(values: Tensor, targets: Tensor) -> Tensor:
    return torch.sum(values * targets) / (
            torch.sqrt(torch.sum(torch.square(values))) * torch.sqrt(torch.sum(torch.square(targets))))


# Angular
def angular_torch(values: Tensor, targets: Tensor) -> Tensor:
    return 2 * torch.arccos(cosine_torch(values, targets)) / np.pi


def loss_torch(name: str) -> Callable[[Tensor, Tensor], Tensor]:
    if name[-4:] == "norm":
        p = int(name[:-5])
        return lambda x, y: p_norm_torch(x, y, p)
    elif name == "mae":
        return lambda x, y: mae_torch(x, y)
    elif name == "mse":
        return lambda x, y: mse_torch(x, y)
    elif name == "rmse":
        return lambda x, y: rmse_torch(x, y)
    elif name == "rmlse":
        return lambda x, y: rmlse_torch(x, y)
    elif name == "lmse":
        return lambda x, y: lmse_torch(x, y)
    elif name == "cosine":
        return lambda x, y: cosine_torch(x, y)
    elif name == "angular":
        return lambda x, y: angular_torch(x, y)
    else:
        raise NotImplementedError(name)


def normal_pdf(x, loc, scale):
    pdf = 1 / (scale * np.sqrt(2 * np.pi)) * torch.exp(-torch.square(x - loc) / (2 * scale ** 2))
    return pdf


def std_normal_cdf(x):
    cdf = 0.5 * (1 + torch.erf(x / np.sqrt(2)))
    return cdf


def skew_normal_pdf(x, shape, loc, scale):
    pdf = 2 * normal_pdf(x, loc, scale) * std_normal_cdf(shape * (x - loc) / scale)
    return pdf


def weibull_pdf(x, shape, scale):
    key = torch.greater_equal(x, 0)
    pdf = torch.zeros_like(x)
    pdf[key] = (shape[key] / scale[key]) * (x[key] / scale[key]) ** (
            shape[key] - 1) * torch.exp(-(x[key] / scale[key]) ** shape[key])
    return pdf


def general_weibull_pdf(x, shape, loc, scale):
    y = x - loc
    key = torch.greater_equal(y, 0)
    pdf = torch.zeros_like(y)
    pdf[key] = (shape[key] / scale[key]) * (y[key] / scale[key]) ** (
            shape[key] - 1) * torch.exp(-(y[key] / scale[key]) ** shape[key])
    return pdf


class NonparametricKernel(torch.nn.Module):
    N_PARAMETERS = None

    def __init__(self, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
        super().__init__()
        self.N_PARAMETERS = n_classes
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if parameters is None:
            self._params = torch.nn.Parameter(torch.randn(n_samples, n_members, n_classes) + 2.0, requires_grad=True)
        else:
            assert parameters.ndim == 2
            assert parameters.shape == (n_members, n_classes)
            self._params = torch.nn.Parameter(
                torch.from_numpy(np.expand_dims(parameters, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        end_members = torch.softmax(self._params, dim=-1)
        return end_members


class NormalKernel(torch.nn.Module):
    N_PARAMETERS = 2

    def __init__(self, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if parameters is None:
            self._params = torch.nn.Parameter(
                torch.randn(n_samples, self.N_PARAMETERS, n_members) + 5.0, requires_grad=True)
        else:
            assert parameters.ndim == 2
            assert parameters.shape == (self.N_PARAMETERS, n_members)
            self._params = torch.nn.Parameter(
                torch.from_numpy(np.expand_dims(parameters, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        locations = self._params[:, 0, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self._params[:, 1, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = normal_pdf(classes, locations, scales) * interval
        return end_members


class SkewNormalKernel(torch.nn.Module):
    N_PARAMETERS = 3

    def __init__(self, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if parameters is None:
            self._params = torch.nn.Parameter(
                torch.randn(n_samples, self.N_PARAMETERS, n_members) + 5.0, requires_grad=True)
        else:
            assert parameters.ndim == 2
            assert parameters.shape == (self.N_PARAMETERS, n_members)
            self._params = torch.nn.Parameter(
                torch.from_numpy(np.expand_dims(parameters, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = self._params[:, 0, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        locations = self._params[:, 1, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self._params[:, 2, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = skew_normal_pdf(classes, shapes, locations, scales) * interval
        return end_members


class WeibullKernel(torch.nn.Module):
    N_PARAMETERS = 2

    def __init__(self, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if parameters is None:
            self._params = torch.nn.Parameter(
                torch.randn(n_samples, self.N_PARAMETERS, n_members) + 5.0, requires_grad=True)
        else:
            assert parameters.ndim == 2
            assert parameters.shape == (self.N_PARAMETERS, n_members)
            self._params = torch.nn.Parameter(
                torch.from_numpy(np.expand_dims(parameters, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = (torch.relu(self._params[:, 0, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self._params[:, 1, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = weibull_pdf(classes, shapes, scales) * interval
        return end_members


class GeneralWeibullKernel(torch.nn.Module):
    N_PARAMETERS = 3

    def __init__(self, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes

        if parameters is None:
            self._params = torch.nn.Parameter(
                torch.randn(n_samples, self.N_PARAMETERS, n_members) + 5.0, requires_grad=True)
        else:
            assert parameters.ndim == 2
            assert parameters.shape == (self.N_PARAMETERS, n_members)
            self._params = torch.nn.Parameter(
                torch.from_numpy(np.expand_dims(parameters, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = (torch.relu(self._params[:, 0, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        locations = self._params[:, 1, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self._params[:, 2, :]) + _INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = general_weibull_pdf(classes, shapes, locations, scales) * interval
        return end_members


class ProportionModule(torch.nn.Module):
    def __init__(self, n_samples: int, n_members: int):
        super().__init__()
        self._params = torch.nn.Parameter(torch.rand(n_samples, 1, n_members), requires_grad=True)

    def forward(self):
        # n_samples x 1 x n_members
        proportions = torch.softmax(self._params, 2)
        return proportions


def get_kernel(kernel_type: KernelType, n_samples: int, n_members: int, n_classes: int, parameters: np.ndarray = None):
    if kernel_type == KernelType.Nonparametric:
        return NonparametricKernel(n_samples, n_members, n_classes, parameters)
    elif kernel_type == KernelType.Normal:
        return NormalKernel(n_samples, n_members, n_classes, parameters)
    elif kernel_type == KernelType.SkewNormal:
        return SkewNormalKernel(n_samples, n_members, n_classes, parameters)
    elif kernel_type == KernelType.Weibull:
        return NormalKernel(n_samples, n_members, n_classes, parameters)
    elif kernel_type == KernelType.GeneralWeibull:
        return GeneralWeibullKernel(n_samples, n_members, n_classes, parameters)
    else:
        raise NotImplementedError(kernel_type)



__all__ = ["PCAAnalyzer"]

from PySide6 import QtCore, QtWidgets

from ..models import Dataset
from ..charts.PCAResultChart import PCAResultChart


class PCAAnalyzer(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(self.tr("PCA Analyzer"))
        self.main_layout = QtWidgets.QGridLayout(self)
        self.chart = PCAResultChart()
        self.main_layout.addWidget(self.chart, 0, 0)
        self.file_dialog = QtWidgets.QFileDialog(self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self._dataset = None

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: Dataset):
        self._dataset = dataset
        self.chart.show_dataset(self._dataset)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("PCA Analyzer"))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import qgrain_pb2 as qgrain__pb2


class QGrainStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.get_service_state = channel.unary_unary(
                '/QGrain/get_service_state',
                request_serializer=qgrain__pb2.ServiceStateRequest.SerializeToString,
                response_deserializer=qgrain__pb2.ServiceStateResponse.FromString,
                )
        self.get_statistics = channel.unary_unary(
                '/QGrain/get_statistics',
                request_serializer=qgrain__pb2.StatisticalRequest.SerializeToString,
                response_deserializer=qgrain__pb2.StatisticalResponse.FromString,
                )
        self.get_ssu_result = channel.unary_unary(
                '/QGrain/get_ssu_result',
                request_serializer=qgrain__pb2.SSURequest.SerializeToString,
                response_deserializer=qgrain__pb2.SSUResponse.FromString,
                )
        self.get_emma_result = channel.unary_unary(
                '/QGrain/get_emma_result',
                request_serializer=qgrain__pb2.EMMARequest.SerializeToString,
                response_deserializer=qgrain__pb2.EMMAResponse.FromString,
                )
        self.get_udm_result = channel.unary_unary(
                '/QGrain/get_udm_result',
                request_serializer=qgrain__pb2.UDMRequest.SerializeToString,
                response_deserializer=qgrain__pb2.UDMResponse.FromString,
                )


class QGrainServicer(object):
    """Missing associated documentation comment in .proto file."""

    def get_service_state(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_statistics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_ssu_result(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_emma_result(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_udm_result(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_QGrainServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'get_service_state': grpc.unary_unary_rpc_method_handler(
                    servicer.get_service_state,
                    request_deserializer=qgrain__pb2.ServiceStateRequest.FromString,
                    response_serializer=qgrain__pb2.ServiceStateResponse.SerializeToString,
            ),
            'get_statistics': grpc.unary_unary_rpc_method_handler(
                    servicer.get_statistics,
                    request_deserializer=qgrain__pb2.StatisticalRequest.FromString,
                    response_serializer=qgrain__pb2.StatisticalResponse.SerializeToString,
            ),
            'get_ssu_result': grpc.unary_unary_rpc_method_handler(
                    servicer.get_ssu_result,
                    request_deserializer=qgrain__pb2.SSURequest.FromString,
                    response_serializer=qgrain__pb2.SSUResponse.SerializeToString,
            ),
            'get_emma_result': grpc.unary_unary_rpc_method_handler(
                    servicer.get_emma_result,
                    request_deserializer=qgrain__pb2.EMMARequest.FromString,
                    response_serializer=qgrain__pb2.EMMAResponse.SerializeToString,
            ),
            'get_udm_result': grpc.unary_unary_rpc_method_handler(
                    servicer.get_udm_result,
                    request_deserializer=qgrain__pb2.UDMRequest.FromString,
                    response_serializer=qgrain__pb2.UDMResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'QGrain', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class QGrain(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def get_service_state(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/QGrain/get_service_state',
            qgrain__pb2.ServiceStateRequest.SerializeToString,
            qgrain__pb2.ServiceStateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_statistics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/QGrain/get_statistics',
            qgrain__pb2.StatisticalRequest.SerializeToString,
            qgrain__pb2.StatisticalResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_ssu_result(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/QGrain/get_ssu_result',
            qgrain__pb2.SSURequest.SerializeToString,
            qgrain__pb2.SSUResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_emma_result(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/QGrain/get_emma_result',
            qgrain__pb2.EMMARequest.SerializeToString,
            qgrain__pb2.EMMAResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_udm_result(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/QGrain/get_udm_result',
            qgrain__pb2.UDMRequest.SerializeToString,
            qgrain__pb2.UDMResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



__all__ = ["SSUResultComponent", "SSUResult"]

import copy
from typing import *

import numpy as np
from numpy import ndarray

from ..statistics import interval_phi
from ..models import DistributionType, ArtificialSample, Sample
from ..distributions import get_distribution
from ..metrics import loss_numpy


class SSUResultComponent:
    __slots__ = ("_classes", "_classes_phi", "_distribution", "_proportion", "_moments", "_parameters")

    def __init__(self, classes: ndarray, classes_phi: ndarray, distribution: ndarray,
                 proportion: float, moments: Tuple[float, float, float, float]):
        self._classes = classes
        self._classes_phi = classes_phi
        self._distribution = distribution
        self._proportion = proportion
        m, std, s, k = moments
        self._moments = dict(mean=m, std=std, skewness=s, kurtosis=k)

    def __repr__(self):
        return f"C({self._moments['mean']:.2f}, {self._proportion:.2%})"

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self._classes_phi)

    @property
    def distribution(self) -> ndarray:
        return self._distribution

    @property
    def proportion(self) -> float:
        return self._proportion

    @property
    def moments(self) -> dict:
        return self._moments

    @property
    def mean(self) -> float:
        return self._moments["mean"]

    @property
    def sorting_coefficient(self) -> float:
        return self._moments["std"]

    @property
    def skewness(self) -> float:
        return self._moments["skewness"]

    @property
    def kurtosis(self) -> float:
        return self._moments["kurtosis"]


class SSUResult:
    """
    This class represents the SSU result of each sample.
    """

    def __init__(self, sample: Union[ArtificialSample, Sample], distribution_type: DistributionType,
                 parameters: ndarray, time_spent: Union[int, float], x0: ndarray = None,
                 settings: Dict[str, Any] = None):
        assert isinstance(sample, (ArtificialSample, Sample))
        assert isinstance(distribution_type, DistributionType)
        assert isinstance(parameters, ndarray)
        assert isinstance(time_spent, (int, float))
        if x0 is not None:
            assert isinstance(x0, ndarray)
            assert x0.ndim == 2
        assert parameters.ndim == 3
        n_iterations, n_parameters, n_components = parameters.shape
        distribution_class = get_distribution(distribution_type)
        assert n_parameters == distribution_class.N_PARAMETERS + 1
        assert n_iterations > 0
        assert n_parameters == 3 or n_parameters == 4
        assert n_components > 0
        self._sample = sample
        self._distribution_type = distribution_type
        self._x0 = x0
        self._parameters = parameters
        self._time_spent = time_spent
        self._settings = settings
        classes = np.expand_dims(np.expand_dims(sample.classes_phi, 0), 0).repeat(n_components, 1)
        proportions, components, (m, std, s, k) = distribution_class.interpret(
            np.expand_dims(self._parameters[-1], 0), classes, self._sample.interval_phi)
        proportions, components, (m, std, s, k) = proportions[0], components[0], (m[0], std[0], s[0], k[0])
        distribution = (proportions @ components)[0]
        proportions = proportions[0]
        self._distribution = distribution
        self._proportions = proportions
        self._components = components
        self._moments = (m, std, s, k)

    def __repr__(self):
        return f"SSUResult({self._sample.name}, {self._parameters.shape[2]}, {self._distribution_type.name})"

    def __len__(self):
        return len(self._components)

    def __iter__(self):
        for i in range(len(self._components)):
            yield self._get_component(i)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_component(item)
        elif isinstance(item, slice):
            return [self._get_component(index) for index in np.arange(len(self._components))[item]]
        else:
            raise TypeError(f"Component indices must be integers or slices, not {type(item)}.")

    @property
    def name(self) -> str:
        return self._sample.name

    @property
    def sample(self) -> Sample:
        return self._sample

    @property
    def classes(self) -> ndarray:
        return self._sample.classes

    @property
    def classes_phi(self) -> ndarray:
        return self._sample.classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self._sample.classes_phi)

    @property
    def distribution(self) -> ndarray:
        return self._distribution

    @property
    def distribution_type(self) -> DistributionType:
        return self._distribution_type

    @property
    def x0(self) -> ndarray:
        return self._x0

    @property
    def parameters(self) -> ndarray:
        return self._parameters

    @property
    def time_spent(self):
        return self._time_spent

    @property
    def n_iterations(self):
        return self._parameters.shape[0]

    @property
    def n_parameters(self) -> int:
        return self._parameters.shape[1]

    @property
    def settings(self) -> Dict[str, Any]:
        return copy.deepcopy(self._settings)

    @property
    def is_valid(self) -> bool:
        valid = True
        for values in [self._proportions, self._components, self._distribution, *self._moments]:
            if np.any(np.logical_or(np.isnan(values), np.isinf(values))):
                valid = False
                break
        return valid

    @property
    def history(self):
        n_iterations, n_parameters, n_components = self._parameters.shape
        classes = np.expand_dims(np.expand_dims(
            self._sample.classes_phi, 0), 0).repeat(n_iterations, 0).repeat(n_components, 1)
        distribution_class = get_distribution(self._distribution_type)
        proportions, components, (m, std, s, k) = distribution_class.interpret(
            self._parameters, classes, self._sample.interval_phi)
        distributions = (proportions @ components)
        for i in range(n_iterations):
            copy_result = copy.copy(self)
            copy_result._distribution = distributions[i, 0]
            copy_result._proportions = proportions[i, 0]
            copy_result._components = components[i]
            copy_result._moments = (m[i], std[i], s[i], k[i])
            yield copy_result

    def _get_component(self, index: int):
        m, std, s, k = self._moments
        component = SSUResultComponent(
            self._sample.classes, self._sample.classes_phi,
            self._components[index], self._proportions[index],
            (m[index], std[index], s[index], k[index]))
        return component

    def loss(self, name: str):
        loss_func = loss_numpy(name)
        return loss_func(self.distribution, self.sample.distribution, None)

    def loss_series(self, name: str):
        n_iterations, n_parameters, n_components = self._parameters.shape
        classes = np.expand_dims(np.expand_dims(
            self._sample.classes_phi, 0), 0).repeat(n_iterations, 0).repeat(n_components, 1)
        distribution_class = get_distribution(self._distribution_type)
        proportions, components, _ = distribution_class.interpret(
            self._parameters, classes, self._sample.interval_phi)
        distributions = (proportions @ components)[:, 0, :]
        targets = np.expand_dims(self._sample.distribution, 0).repeat(n_iterations, 0)
        loss_func = loss_numpy(name)
        loss_series = loss_func(distributions, targets, 1)
        return loss_series



import logging
import string
from concurrent import futures

import grpc
import numpy as np

from ..statistics import to_phi, all_statistics
from ..models import DistributionType, KernelType, Sample, SSUResult, validate_distributions, validate_classes, Dataset
from ..distributions import get_distribution
from ..ssu import try_ssu
from ..emma import try_emma
from ..udm import try_udm
from . import qgrain_pb2 as qgrain_pb2
from . import qgrain_pb2_grpc as qgrain_pb2_grpc


class QGrainServicer(qgrain_pb2_grpc.QGrainServicer):
    logger = logging.getLogger("QGrain.QGrainServicer")

    def __init__(self, address: str = "localhost:50051", max_workers: int = 8, max_message_length: int = 2**30,
                 max_dataset_size: int = 100000):
        super(QGrainServicer, self).__init__()
        self._address = address
        self._max_workers = max_workers
        self._max_message_length = max_message_length
        self._max_dataset_size = max_dataset_size
    
    @classmethod
    def _to_sample(cls, sample_pb2) -> Sample:
        if len(sample_pb2.classes) == 0:
            raise ValueError("The series of grain size classes is empty.")
        if len(sample_pb2.distribution) == 0:
            raise ValueError("The series of grain size distribution is empty.")
        if len(sample_pb2.classes) != len(sample_pb2.distribution):
            raise ValueError("The lengths of grain size classes and frequencies are not equal.")
        valid, array_or_msg = validate_classes(sample_pb2.classes)
        if not valid:
            raise ValueError(array_or_msg)
        classes = array_or_msg.astype(np.float64)
        classes_phi = to_phi(classes)
        valid, array_or_msg = validate_distributions([sample_pb2.distribution])
        if not valid:
            raise ValueError(array_or_msg)
        distribution = array_or_msg.astype(np.float64)[0]
        sample = Sample(sample_pb2.name, classes, classes_phi, distribution)
        return sample

    @classmethod
    def _to_dataset(cls, dataset_pb2: qgrain_pb2.Dataset) -> Dataset:
        sample_names = list(dataset_pb2.sample_names)
        classes = np.array(dataset_pb2.classes, dtype=np.float64)
        shape = (dataset_pb2.n_samples, dataset_pb2.n_classes)
        distributions = np.frombuffer(dataset_pb2.distributions, dtype=np.float64).copy().reshape(shape)
        dataset = Dataset(dataset_pb2.name, sample_names, classes, distributions)
        return dataset

    def get_service_state(self, request: qgrain_pb2.ServiceStateRequest, context):
        devices = ["cpu"]
        import torch
        if torch.cuda.is_available():
            devices.append("cuda")
        devices.extend([f"cuda:{i}" for i in range(torch.cuda.device_count())])
        if "cuda:0" in devices:
            devices.remove("cuda:0")
        state = dict(max_workers=self._max_workers, max_message_length=self._max_message_length,
                     max_dataset_size=self._max_dataset_size, available_devices=tuple(devices))
        response = qgrain_pb2.ServiceStateResponse(**state)
        return response

    def get_statistics(self, request: qgrain_pb2.StatisticalRequest, context):
        try:
            dataset = self._to_dataset(request.dataset)
        except TypeError as e:
            return qgrain_pb2.StatisticalResponse(message=f"The dataset in request is invalid, please check: {e}.")
        except ValueError as e:
            return qgrain_pb2.StatisticalResponse(message=f"The dataset in request is invalid, please check: {e}.")
        results = []
        for sample in dataset:
            all_parameters = all_statistics(sample.classes, sample.classes_phi, sample.distribution)
            methods = ("arithmetic", "geometric", "logarithmic", "geometric_fw57", "logarithmic_fw57")
            parameters_of_methods = {}
            for method in methods:
                parameters = qgrain_pb2.StatisticalParameters(method=method, **all_parameters[method])
                parameters_of_methods[method] = parameters
            other_keys = ("proportions_gsm", "proportions_ssc", "proportions_bgssc",
                          "group_folk54", "group_bp12_symbol", "group_bp12")
            other_parameters = {key: all_parameters[key] for key in other_keys}
            proportions = {string.capwords(f"{adj} {grade}"): proportion for (adj, grade), proportion in
                           all_parameters["proportions"].items()}
            result = qgrain_pb2.StatisticalResult(**parameters_of_methods, **other_parameters, proportions=proportions)
            results.append(result)
        message = f"Success! {len(dataset)} statistical results of dataset {dataset.name} are available."
        response = qgrain_pb2.StatisticalResponse(message=message, results=results)
        return response

    def get_ssu_result(self, request: qgrain_pb2.SSURequest, context):
        try:
            sample = self._to_sample(request.sample)
        except ValueError as e:
            return qgrain_pb2.SSUResponse(message=f"The sample in request is invalid, please check: {e}.")
        distribution_type = {t.value: t for name, t in DistributionType.__members__.items()}[request.distribution_type]
        distribution_class = get_distribution(distribution_type)
        n_parameters = distribution_class.N_PARAMETERS + 1
        if len(request.x0) == 0:
            x0 = None
        else:
            x0 = np.frombuffer(request.x0, dtype=np.float64).copy().reshape(n_parameters, request.n_components)
        settings = dict(loss=request.loss, optimizer=request.optimizer, try_global=request.try_global,
                        global_max_niter=request.global_max_niter, global_niter_success=request.global_niter_success,
                        global_step_size=request.global_step_size, optimizer_max_niter=request.optimizer_max_niter,
                        need_history=request.need_history)
        try:
            result, message = try_ssu(sample, distribution_type, request.n_components, x0=x0,
                                      **settings, logger=self.logger)
        except AssertionError as e:
            return qgrain_pb2.SSUResponse(message=f"The algorithm settings in request is invalid, please check: {e}.")
        if isinstance(result, SSUResult):
            response = qgrain_pb2.SSUResponse(
                message=f"Success! The SSU result of sample {sample.name} is available.", time_spent=result.time_spent,
                n_iterations=result.n_iterations, n_parameters=result.n_parameters, n_components=len(result),
                parameters=result.parameters.astype(np.float64).tobytes())
            return response
        else:
            response = qgrain_pb2.SSUResponse(message=message)
            return response

    def get_emma_result(self, request: qgrain_pb2.EMMARequest, context):
        try:
            dataset = self._to_dataset(request.dataset)
        except TypeError as e:
            return qgrain_pb2.EMMAResponse(message=f"The dataset in request is invalid, please check: {e}.")
        except ValueError as e:
            return qgrain_pb2.EMMAResponse(message=f"The dataset in request is invalid, please check: {e}.")
        distribution_type = {t.value: t for name, t in DistributionType.__members__.items()}[request.distribution_type]
        kernel_type = {t.value: t for name, t in KernelType.__members__.items()}[request.distribution_type]
        n_parameters = get_distribution(distribution_type).N_PARAMETERS
        if len(request.x0) == 0:
            x0 = None
        else:
            x0 = np.frombuffer(request.x0, dtype=np.float64).copy().reshape(n_parameters, request.n_members)
        settings = dict(device=request.device, loss=request.loss, pretrain_epochs=request.pretrain_epochs,
                        min_epochs=request.min_epochs, max_epochs=request.max_epochs, precision=request.precision,
                        learning_rate=request.learning_rate, betas=tuple(request.betas),
                        update_end_members=request.update_end_members, need_history=request.need_history)
        try:
            result = try_emma(dataset, kernel_type, request.n_members, x0=x0, **settings, logger=self.logger)
        except AssertionError as e:
            return qgrain_pb2.EMMAResponse(message=f"The algorithm settings in request is invalid, please check: {e}.")
        response = qgrain_pb2.EMMAResponse(
            message=f"Success! The EMMA result of dataset {dataset.name} is available.", time_spent=result.time_spent,
            n_iterations=result.n_iterations, n_samples=result.n_samples, n_members=result.n_members,
            n_classes=result.n_classes, proportions=result._proportions.astype(np.float64).tobytes(),
            end_members=result._end_members.astype(np.float64).tobytes(), losses=result.loss_series(request.loss))
        return response

    def get_udm_result(self, request: qgrain_pb2.UDMRequest, context):
        try:
            dataset = self._to_dataset(request.dataset)
        except TypeError as e:
            return qgrain_pb2.UDMResponse(message=f"The dataset in request is invalid, please check: {e}.")
        except ValueError as e:
            return qgrain_pb2.UDMResponse(message=f"The dataset in request is invalid, please check: {e}.")
        distribution_type = {t.value: t for name, t in DistributionType.__members__.items()}[request.distribution_type]
        kernel_type = {t.value: t for name, t in KernelType.__members__.items()}[request.distribution_type]
        n_parameters = get_distribution(distribution_type).N_PARAMETERS
        if len(request.x0) == 0:
            x0 = None
        else:
            x0 = np.frombuffer(request.x0, dtype=np.float64).copy().reshape(n_parameters, request.n_components)
        settings = dict(device=request.device, pretrain_epochs=request.pretrain_epochs, min_epochs=request.min_epochs,
                        max_epochs=request.max_epochs, precision=request.precision, learning_rate=request.learning_rate,
                        betas=tuple(request.betas), constraint_level=request.constraint_level,
                        need_history=request.need_history)
        try:
            result = try_udm(dataset, kernel_type, request.n_components, x0=x0, **settings, logger=self.logger)
        except AssertionError as e:
            return qgrain_pb2.UDMResponse(
                message=f"The algorithm settings in request is invalid, please check: {e}.")
        response = qgrain_pb2.UDMResponse(
            message=f"Success! The UDM result of dataset {dataset.name} is available.", time_spent=result.time_spent,
            n_iterations=result.n_iterations, n_samples=result.n_samples, n_components=result.n_components,
            n_classes=result.n_classes, parameters=result.parameters.astype(np.float64).tobytes(),
            distribution_losses=result.loss_series("distribution"), component_losses=result.loss_series("component"),
            total_losses=result.loss_series("total"))
        return response

    def serve(self):
        self.logger.info(f"Starting the server and listening the address: {self._address}. "
                         f"Max thread workers is {self._max_workers}. "
                         f"Max length of the message is {self._max_message_length} bytes. "
                         f"Max size of the grain size dataset is {self._max_dataset_size} samples.\n")
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=self._max_workers),
                             options=[("grpc.max_send_message_length", self._max_message_length),
                                      ("grpc.max_receive_message_length", self._max_message_length)])
        qgrain_pb2_grpc.add_QGrainServicer_to_server(QGrainServicer(), server)
        server.add_insecure_port(self._address)
        server.start()
        server.wait_for_termination()



__all__ = ["ParameterEditor"]

from typing import *

import numpy as np
from PySide6 import QtCore, QtWidgets
from numpy import ndarray

from ..statistics import interval_phi
from ..models import DistributionType, SSUResult, ArtificialDataset
from ..distributions import sort_parameters
from ..charts.DistributionChart import DistributionChart


class ParameterComponent(QtWidgets.QWidget):
    NORMAL_SETTINGS = dict(
        n_parameters=3,
        parameter_names=("Location", "Scale", "Weight"),
        ranges=((-15.0, 15.0), (0.01, 100.0), (0.0, 10.0)),
        defaults=(5.0, 1.0, 1.0),
        steps=(0.1, 0.1, 0.1))

    SKEW_NORMAL_SETTINGS = dict(
        n_parameters=4,
        parameter_names=("Shape", "Location", "Scale", "Weight"),
        ranges=((-100.0, 100.0), (-15.0, 15.0), (0.01, 100.0), (0.0, 10.0)),
        defaults=(0.0, 5.0, 1.0, 1.0),
        steps=(0.1, 0.1, 0.1, 0.1))

    WEIBULL_SETTINGS = dict(
        n_parameters=3,
        parameter_names=("Shape", "Scale", "Weight"),
        ranges=((-2000.0, 2000.0), (0.01, 2000.0), (0.0, 10.0)),
        defaults=(3.6, 1.0, 1.0),
        steps=(0.1, 0.1, 0.1))

    GENERAL_WEIBULL_SETTINGS = dict(
        n_parameters=4,
        parameter_names=("Shape", "Location", "Scale", "Weight"),
        ranges=((-2000.0, 2000.0), (-2000.0, 2000.0), (0.01, 2000.0), (0.0, 10.0)),
        defaults=(3.6, 5.0, 1.0, 1.0),
        steps=(0.1, 0.1, 0.1, 0.1))

    SETTING_MAP = {
        DistributionType.Normal: NORMAL_SETTINGS,
        DistributionType.SkewNormal: SKEW_NORMAL_SETTINGS,
        DistributionType.Weibull: WEIBULL_SETTINGS,
        DistributionType.GeneralWeibull: GENERAL_WEIBULL_SETTINGS}

    value_changed = QtCore.Signal()

    def __init__(self, name: str, distribution_type: DistributionType, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self._name = name
        self._distribution_type = distribution_type
        # make sure that the pyside2-lupdate.exe can recognize these keywords
        _ = [self.tr("Shape"), self.tr("Location"), self.tr("Scale"), self.tr("Weight")]
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.group = QtWidgets.QGroupBox("")
        self.group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.main_layout.addWidget(self.group)
        self.group_layout = QtWidgets.QGridLayout(self.group)
        self.group_layout.setColumnMinimumWidth(0, 32)
        self.group_layout.setColumnMinimumWidth(1, 32)
        settings = self.SETTING_MAP[self._distribution_type]
        self.widgets = [] # type: list[tuple[QtWidgets.QLabel, QtWidgets.QDoubleSpinBox]]
        self.name_label = QtWidgets.QLabel(self._name)
        self.name_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.value_label = QtWidgets.QLabel(self.tr("Value"))
        self.value_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.group_layout.addWidget(self.name_label, 0, 0)
        self.group_layout.addWidget(self.value_label, 0, 1)
        for i in range(settings["n_parameters"]):
            label = QtWidgets.QLabel(self.tr(settings["parameter_names"][i]))
            label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
            input = QtWidgets.QDoubleSpinBox()
            input.setRange(*settings["ranges"][i])
            input.setSingleStep(settings["steps"][i])
            input.setValue(settings["defaults"][i])
            self.group_layout.addWidget(label, i+1, 0)
            self.group_layout.addWidget(input, i+1, 1)
            self.widgets.append((label, input))
        for _, input in self.widgets:
            input.valueChanged.connect(self.on_value_changed)

    @property
    def parameters(self):
        return [input.value() for _, input in self.widgets]

    def set_parameters(self, parameters):
        self.blockSignals(True)
        for (_, input), value in zip(self.widgets, parameters):
            input.setValue(value)
        self.blockSignals(False)

    def on_value_changed(self, _):
        self.value_changed.emit()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.value_label.setText(self.tr("Value"))
        settings = self.SETTING_MAP[self._distribution_type]
        for param_name, (param_label, _) in zip(settings["parameter_names"], self.widgets):
            param_label.setText(self.tr(param_name))


class ParameterEditor(QtWidgets.QDialog):
    SUPPORT_DISTRIBUTIONS = (
        (DistributionType.Normal, "Normal"),
        (DistributionType.SkewNormal, "Skew Normal"),
        (DistributionType.Weibull, "Weibull"),
        (DistributionType.GeneralWeibull, "General Weibull"))
    DISTRIBUTION_INDEX_MAP = {
        DistributionType.Normal: 0,
        DistributionType.SkewNormal: 1,
        DistributionType.Weibull: 2,
        DistributionType.GeneralWeibull: 3}

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._cache_dict = {}
        self._classes = np.logspace(0, 5, 101) * 0.02
        self._classes_phi = -np.log2(self._classes / 1000.0)
        self._interval_phi = interval_phi(self._classes_phi)
        self._target = None
        self.setWindowTitle(self.tr("Parameter Editor"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.preset_label = QtWidgets.QLabel(self.tr("Preset Archive"))
        self.preset_combo_box = QtWidgets.QComboBox()
        self.preset_combo_box.addItems([str(i) for i in range(1, 11)])
        self.preset_combo_box.currentIndexChanged.connect(self.switch_preset)
        self.enabled_checkbox = QtWidgets.QCheckBox()
        self.enabled_checkbox.setText(self.tr("Enabled"))
        self.preview_button = QtWidgets.QPushButton(self.tr("Preview"))
        self.preview_button.clicked.connect(self.update_chart)
        self.control_layout.addWidget(self.preset_label, 0, 0)
        self.control_layout.addWidget(self.preset_combo_box, 0, 1)
        self.control_layout.addWidget(self.enabled_checkbox, 0, 2)
        self.control_layout.addWidget(self.preview_button, 0, 3)
        self.n_components_label = QtWidgets.QLabel(self.tr("Number of Components"))
        self.n_components_input = QtWidgets.QSpinBox()
        self.n_components_input.setRange(1, 12)
        self.n_components_input.valueChanged.connect(self.on_n_components_changed)
        self.n_components_input.valueChanged.connect(self.update_cache)
        self.distribution_type_label = QtWidgets.QLabel(self.tr("Distribution Type"))
        self.distribution_type_combo_box = QtWidgets.QComboBox()
        self.distribution_type_combo_box.addItems([name for _, name in self.SUPPORT_DISTRIBUTIONS])
        self.distribution_type_combo_box.currentIndexChanged.connect(self.on_distribution_type_changed)
        self.distribution_type_combo_box.currentIndexChanged.connect(self.on_distribution_type_changed)
        self.control_layout.addWidget(self.n_components_label, 1, 0)
        self.control_layout.addWidget(self.n_components_input, 1, 1)
        self.control_layout.addWidget(self.distribution_type_label, 1, 2)
        self.control_layout.addWidget(self.distribution_type_combo_box, 1, 3)
        self.parameter_tab_widget = QtWidgets.QTabWidget()
        self.component_sets: List[Tuple[QtWidgets.QWidget, QtWidgets.QGridLayout, List[ParameterComponent]]] = []
        self.preview_group = QtWidgets.QGroupBox(self.tr("Preview"))
        self.preview_layout = QtWidgets.QGridLayout(self.preview_group)
        self.preview_layout.setContentsMargins(0, 0, 0, 0)
        self.preview_chart = DistributionChart(parent=self)
        self.preview_layout.addWidget(self.preview_chart, 0, 0)
        self.splitter_1 = QtWidgets.QSplitter()
        self.splitter_1.setOrientation(QtCore.Qt.Vertical)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.parameter_tab_widget)
        self.splitter_1.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.splitter_2 = QtWidgets.QSplitter()
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.preview_group)
        self.main_layout.addWidget(self.splitter_2, 0, 0)

        self.switch_preset(0)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

    @property
    def preset_index(self) -> int:
        return self.preset_combo_box.currentIndex()

    @property
    def n_components(self) -> int:
        return self.n_components_input.value()

    @property
    def distribution_type(self) -> DistributionType:
        return self.SUPPORT_DISTRIBUTIONS[self.distribution_type_combo_box.currentIndex()][0]

    @property
    def parameter_enabled(self) -> bool:
        return self.enabled_checkbox.isChecked()

    @property
    def parameters(self) -> np.ndarray:
        parameters = []
        for _, _, components in self.component_sets:
            for component in components:
                if component.isEnabled():
                    parameters.append(component.parameters)
        parameters = np.expand_dims(np.array(parameters).T, axis=0)
        classes = np.expand_dims(np.expand_dims(self._classes_phi, axis=0), axis=0).repeat(self.n_components, axis=1)
        sorted_parameters = sort_parameters(self.distribution_type, parameters, classes, self._interval_phi)
        return sorted_parameters[0]

    def _clear_components(self):
        for param_holder, param_layout, components in self.component_sets:
            for component in components:
                component.value_changed.disconnect(self.update_cache)
                param_layout.removeWidget(component)
                component.hide()
            param_holder.hide()
        self.parameter_tab_widget.clear()
        self.component_sets.clear()

    def _add_components(self):
        for i in range(4):
            param_holder = QtWidgets.QWidget()
            self.parameter_tab_widget.addTab(param_holder, f"C{i * 3 + 1}-{i * 3 + 3}")
            param_layout = QtWidgets.QGridLayout(param_holder)
            param_layout.setContentsMargins(0, 0, 0, 0)
            components = []
            for j in range(3):
                component = ParameterComponent(f"C{i*3+j+1}", self.distribution_type)
                component.value_changed.connect(self.update_cache)
                param_layout.addWidget(component, 0, j)
                components.append(component)
            self.component_sets.append((param_holder, param_layout, components))

    def on_n_components_changed(self, n_components: int):
        for i in range(4):
            for j in range(3):
                current_index = i*3+j
                if current_index < n_components:
                    self.component_sets[i][2][j].setEnabled(True)
                else:
                    self.component_sets[i][2][j].setEnabled(False)

    def on_distribution_type_changed(self, _):
        self._clear_components()
        self._add_components()
        self.on_n_components_changed(self.n_components)

    def update_cache(self):
        self._cache_dict[self.preset_index] = (self.distribution_type, self.n_components, self.parameters)

    def switch_preset(self, preset_index: int):
        if preset_index in self._cache_dict.keys():
            distribution_type, n_components, parameter_matrix = self._cache_dict[preset_index]
            self.distribution_type_combo_box.setCurrentIndex(self.DISTRIBUTION_INDEX_MAP[distribution_type])
            self.n_components_input.setValue(n_components)
            parameters = parameter_matrix.T
        else:
            self.distribution_type_combo_box.setCurrentIndex(3)
            self.n_components_input.setValue(3)
            parameters = [
                [611.76, -467.20, 477.40, 0.81],
                [2.28, 5.35, 2.57, 2.12],
                [2.74, 2.67, 2.68, 2.77]]
        enabled_components = [] # type: list[ParameterComponent]
        for _, _, components in self.component_sets:
            for component in components:
                if component.isEnabled():
                    enabled_components.append(component)
        for component_parameters, component in zip(parameters, enabled_components):
            component.set_parameters(component_parameters)
        self.update_cache()
        self.update_chart()

    def setup_target(self, classes: np.ndarray, target: np.ndarray):
        self._classes = classes
        self._classes_phi = -np.log2(classes / 1000.0)
        self._interval_phi = interval_phi(self._classes_phi)
        self._target = target

    def update_chart(self):
        sample = ArtificialDataset(np.expand_dims(self.parameters, axis=0), self.distribution_type)[0]
        self.preview_chart.show_chart(sample)

    def refer_parameters(self, distribution_type: DistributionType, parameters: ndarray):
        assert isinstance(distribution_type, DistributionType)
        assert isinstance(parameters, ndarray)
        assert parameters.ndim == 2
        n_parameters, n_components = parameters.shape
        self.distribution_type_combo_box.setCurrentIndex(self.DISTRIBUTION_INDEX_MAP[distribution_type])
        self.n_components_input.setValue(n_components)
        enabled_components: List[ParameterComponent] = []
        for _, _, components in self.component_sets:
            for component in components:
                if component.isEnabled():
                    enabled_components.append(component)
        for i, component in enumerate(enabled_components):
            component.set_parameters(parameters[:, i])
        self.update_chart()
        self.update_cache()

    def refer_result(self, result: SSUResult):
        self.distribution_type_combo_box.setCurrentIndex(self.DISTRIBUTION_INDEX_MAP[result.distribution_type])
        self.n_components_input.setValue(len(result))
        enabled_components: List[ParameterComponent] = []
        for _, _, components in self.component_sets:
            for component in components:
                if component.isEnabled():
                    enabled_components.append(component)
        for i, component in enumerate(enabled_components):
            component.set_parameters(result.parameters[-1, :, i])
        self.update_chart()
        self.update_cache()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Parameter Editor"))
        self.control_group.setTitle(self.tr("Control"))
        self.preset_label.setText(self.tr("Preset Archive"))
        self.enabled_checkbox.setText(self.tr("Enabled"))
        self.preview_button.setText(self.tr("Preview"))
        self.n_components_label.setText(self.tr("Number of Components"))
        self.distribution_type_label.setText(self.tr("Distribution Type"))
        self.preview_group.setTitle(self.tr("Preview"))




import multiprocessing as mp
from enum import Enum, unique
from typing import *

import psutil
from PySide6 import QtCore, QtWidgets

from ..models import SSUResult
from ..ssu import try_ssu
from ..charts import background_color


@unique
class TaskState(Enum):
    NotStarted = 0
    Processing = 1
    Failed = 2
    Finished = 3


def execute_tasks(
        task_queue: mp.Queue,
        result_queue: mp.Queue,
        failed_queue: mp.Queue,
        event_queue: mp.Queue,
        retry_times: int = 5):
    while True:
        task_i, task = task_queue.get()
        left_times = retry_times
        event_queue.put((task_i, TaskState.Processing))
        while left_times > 0:
            result, msg = try_ssu(**task)
            if isinstance(result, SSUResult):
                result_queue.put((task_i, result))
                break
            else:
                left_times -= 1
        if left_times > 0:
            event_queue.put((task_i, TaskState.Finished))
        else:
            failed_queue.put((task_i, task))
            event_queue.put((task_i, TaskState.Failed))


class TaskStateBubble(QtWidgets.QWidget):
    def __init__(self, size=16, border_radius=4):
        super().__init__()
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self._size = size
        self._border_radius = border_radius
        self.main_layout = QtWidgets.QGridLayout(self)
        self.display = QtWidgets.QLabel()
        self.display.setFixedSize(self._size, self._size)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.display)
        self.change_state(TaskState.NotStarted)

    def get_qss(self, state: TaskState):
        if state == TaskState.NotStarted:
            return f"border-radius: {self._border_radius}px;border:0px;background:#7a7374"
        elif state == TaskState.Processing:
            return f"border-radius: {self._border_radius}px;border:0px;background:#ff9900"
        elif state == TaskState.Failed:
            return f"border-radius: {self._border_radius}px;border:0px;background:#36292f"
        elif state == TaskState.Finished:
            return f"border-radius: {self._border_radius}px;border:0px;background:#8abcd1"
        else:
            raise NotImplementedError(state)

    def change_state(self, state: TaskState):
        qss = self.get_qss(state)
        self.display.setStyleSheet(qss)

    def make_transparent(self):
        sheet = f"border-radius: {self._border_radius}px;border:0px;background:{background_color()}"
        self.display.setStyleSheet(sheet)


class SSUMulticoreAnalyzer(QtWidgets.QDialog):
    BUBBLE_ROWS = 10
    BUBBLE_COLUMNS = 20
    result_finished = QtCore.Signal(SSUResult)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.__task_queue = mp.Queue()
        self.__result_queue = mp.Queue()
        self.__failed_queue = mp.Queue()
        self.__event_queue = mp.Queue()
        self.__tasks: List[Dict] = []
        self.__state_map: Dict[int, TaskState] = {}
        self.__results: Dict[int, SSUResult] = {}
        self.__failed_tasks: Dict[int, Dict] = {}
        self.__processes: List[mp.Process] = []
        self.__start_flag = False
        self.__watcher = QtCore.QTimer()
        self.__watcher.setSingleShot(False)
        self.__watcher.setInterval(10)
        self.__watcher.timeout.connect(self.watch)

        self.setWindowTitle(self.tr("SSU Multicore Analyzer"))
        self.main_layout = QtWidgets.QGridLayout(self)
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.n_tasks_label = QtWidgets.QLabel(self.tr("Total Number of Tasks"))
        self.n_tasks_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.n_workers_label = QtWidgets.QLabel(self.tr("Number of Workers"))
        self.n_workers_input = QtWidgets.QSpinBox()
        self.n_workers_input.setRange(1, psutil.cpu_count(logical=False))
        self.n_workers_input.setValue(2)
        self.n_remains_label = QtWidgets.QLabel(self.tr("Number of Remaining Tasks"))
        self.n_remains_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.n_finished_tasks_label = QtWidgets.QLabel(self.tr("Number of Finished Tasks"))
        self.n_finished_tasks_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.n_failed_tasks_label = QtWidgets.QLabel(self.tr("Number of Failed Tasks"))
        self.n_failed_tasks_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.run_button = QtWidgets.QPushButton(self.tr("Start"))
        self.run_button.clicked.connect(self.on_run_button_clicked)
        self.control_layout.addWidget(self.n_tasks_label, 0, 0)
        self.control_layout.addWidget(self.n_tasks_display, 0, 1)
        self.control_layout.addWidget(self.n_workers_label, 0, 2)
        self.control_layout.addWidget(self.n_workers_input, 0, 3)
        self.control_layout.addWidget(self.n_remains_label, 1, 0)
        self.control_layout.addWidget(self.n_remains_display, 1, 1)
        self.control_layout.addWidget(self.n_finished_tasks_label, 1, 2)
        self.control_layout.addWidget(self.n_finished_tasks_display, 1, 3)
        self.control_layout.addWidget(self.n_failed_tasks_label, 2, 0)
        self.control_layout.addWidget(self.n_failed_tasks_display, 2, 1)
        self.control_layout.addWidget(self.run_button, 2, 2, 1, 2)

        self.state_group = QtWidgets.QGroupBox(self.tr("State"))
        self.state_group.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.state_layout = QtWidgets.QGridLayout(self.state_group)
        self.__bubbles = [] # type: list[TaskStateBubble]
        self.bubble_holder = QtWidgets.QWidget()
        self.bubble_layout = QtWidgets.QGridLayout(self.bubble_holder)
        for i in range(self.BUBBLE_ROWS):
            for j in range(self.BUBBLE_COLUMNS):
                bubble = TaskStateBubble()
                self.bubble_layout.addWidget(bubble, i, j)
                self.__bubbles.append(bubble)
        self.previous_button = QtWidgets.QPushButton(self.tr("Previous"))
        self.page_combo_box = QtWidgets.QComboBox()
        self.page_combo_box.addItem(self.tr("No Page"))
        self.next_button = QtWidgets.QPushButton(self.tr("Next"))
        self.previous_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(max(self.page_index-1, 0)))
        self.page_combo_box.currentIndexChanged.connect(self.update_page)
        self.next_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(min(self.page_index+1, self.n_pages-1)))
        self.state_layout.addWidget(self.bubble_holder, 0, 0, 1, 3)
        self.state_layout.addWidget(self.previous_button, 1, 0)
        self.state_layout.addWidget(self.page_combo_box, 1, 1)
        self.state_layout.addWidget(self.next_button, 1, 2)

        self.main_layout.addWidget(self.control_group, 0, 0)
        self.main_layout.addWidget(self.state_group, 1, 0)

    @property
    def n_tasks(self) -> int:
        return len(self.__tasks)

    @property
    def page_size(self) -> int:
        return self.BUBBLE_ROWS * self.BUBBLE_COLUMNS

    @property
    def n_pages(self) -> int:
        n_pages, left = divmod(self.n_tasks, self.page_size)
        if left > 0:
            n_pages += 1
        return n_pages

    @property
    def page_index(self) -> int:
        return self.page_combo_box.currentIndex()

    def in_page(self, index: int) -> bool:
        return self.page_size*self.page_index <= index < self.page_size * (self.page_index + 1)

    def setup_processes(self):
        self.__watcher.start()
        n_processes = self.n_workers_input.value()
        for i in range(n_processes):
            process = mp.Process(
                target=execute_tasks,
                args=(self.__task_queue,
                      self.__result_queue,
                      self.__failed_queue,
                      self.__event_queue))
            process.start()
            self.__processes.append(process)

    def kill_processes(self):
        self.__watcher.stop()
        for process in self.__processes:
            process.kill()
            process.join()
        self.__processes.clear()
        self.__task_queue.close()
        self.__failed_queue.close()
        self.__result_queue.close()
        self.__event_queue.close()
        self.__task_queue = mp.Queue()
        self.__result_queue = mp.Queue()
        self.__failed_queue = mp.Queue()
        self.__event_queue = mp.Queue()

    def setup_tasks(self, tasks: List[Dict]):
        self.kill_processes()
        self.__tasks.clear()
        self.__state_map.clear()
        self.__results.clear()
        self.__failed_tasks.clear()
        for i, task in enumerate(tasks):
            self.__state_map[i] = TaskState.NotStarted
            self.__tasks.append(task)
            self.__task_queue.put((i, task))
        self.n_tasks_display.setText(str(self.n_tasks))
        self.n_remains_display.setText(str(self.n_tasks))
        self.n_failed_tasks_display.setText("0")
        self.n_finished_tasks_display.setText("0")
        self.page_combo_box.clear()
        for i in range(self.n_pages):
            self.page_combo_box.addItem(self.tr("Page {0}").format(i+1))
        self.page_combo_box.setCurrentIndex(0)
        self.n_workers_input.setEnabled(True)
        self.run_button.setEnabled(True)

    def on_run_button_clicked(self):
        self.n_workers_input.setEnabled(False)
        self.run_button.setEnabled(False)
        self.setup_processes()

    def update_page(self):
        if self.page_index < 0:
            return
        offset = self.page_index * self.page_size
        for i in range(self.page_size):
            index = i + offset
            bubble = self.__bubbles[i]
            if index >= self.n_tasks:
                bubble.make_transparent()
                bubble.setToolTip("")
            else:
                state = self.__state_map[index]
                task = self.__tasks[index]
                bubble.show()
                bubble.change_state(state)
                bubble.setToolTip(task["sample"].name)

    def watch(self):
        try:
            while True:
                event: Tuple[int, TaskState] = self.__event_queue.get(block=False)
                task_i, state = event
                self.__state_map[task_i] = state
                if self.in_page(task_i):
                    offset = self.page_index*self.page_size
                    bubble = self.__bubbles[task_i - offset]
                    bubble.change_state(state)
        except Exception:
            pass

        try:
            while True:
                task_i, result = self.__result_queue.get(block=False)
                self.__results[task_i] = result
        except Exception:
            pass

        try:
            while True:
                task_i, task = self.__failed_queue.get(block=False)
                self.__failed_tasks[task_i] = task
        except Exception:
            pass

        n_remains = 0
        n_failed = 0
        n_finished = 0
        for index, state in self.__state_map.items():
            if state == TaskState.NotStarted or state == TaskState.Processing:
                n_remains += 1
            elif state == TaskState.Failed:
                n_failed += 1
            else:
                n_finished += 1
        self.n_remains_display.setText(str(n_remains))
        self.n_failed_tasks_display.setText(str(n_failed))
        self.n_finished_tasks_display.setText(str(n_finished))

        if n_remains == 0 and n_failed == len(self.__failed_tasks) and n_finished == len(self.__results):
            results = [(index, result) for index, result in self.__results.items()]
            results.sort(key=lambda x: x[0])
            for index, result in results:
                self.result_finished.emit(result)
            self.kill_processes()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Multicore Analyzer"))
        self.control_group.setTitle(self.tr("Control"))
        self.n_tasks_label.setText(self.tr("Total Number of Tasks"))
        self.n_workers_label.setText(self.tr("Number of Workers"))
        self.n_remains_label.setText(self.tr("Number of Remaining Tasks"))
        self.n_finished_tasks_label.setText(self.tr("Number of Finished Tasks"))
        self.n_failed_tasks_label.setText(self.tr("Number of Failed Tasks"))
        if self.n_tasks == 0:
            self.n_tasks_display.setText(self.tr("Unknown"))
            self.n_remains_display.setText(self.tr("Unknown"))
            self.n_finished_tasks_display.setText(self.tr("Unknown"))
            self.n_failed_tasks_display.setText(self.tr("Unknown"))
            self.page_combo_box.addItem(self.tr("No Page"))
        else:
            for i in range(self.page_combo_box.count()):
                self.page_combo_box.setItemText(i, self.tr("Page {0}").format(i+1))
        self.run_button.setText(self.tr("Start"))
        self.state_group.setTitle(self.tr("State"))
        self.previous_button.setText(self.tr("Previous"))
        self.next_button.setText(self.tr("Next"))



import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler

import matplotlib as mpl
from PySide6 import QtWidgets, QtCore, QtGui

from .RuntimeLog import StatusBarLogHandler, GUILogHandler, RuntimeLog
from .. import QGRAIN_VERSION, QGRAIN_ROOT_PATH
from ..charts import setup_matplotlib

EXTRA = {"font_family": "Roboto, Arial, Helvetica, Tahoma, Verdana, Microsoft YaHei UI, SimSum",
         "density_scale": "-1"}


def get_dir_size(directory: str):
    size = 0
    for root, _, files in os.walk(directory):
        size += sum([os.path.getsize(os.path.join(root, name)) for name in files])
    return size


def create_necessary_folders():
    necessary_folders = (
        os.path.join(os.path.expanduser("~"), "QGrain"),
        os.path.join(os.path.expanduser("~"), "QGrain", "logs"))
    for folder in necessary_folders:
        os.makedirs(folder, exist_ok=True)


def setup_language(app: QtWidgets.QApplication, language: str):
    trans = QtCore.QTranslator(app)
    trans.load(os.path.join(QGRAIN_ROOT_PATH, "assets", language))
    app.installTranslator(trans)


def setup_logging(status_bar: QtWidgets.QStatusBar, log_dialog: RuntimeLog):
    format_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    file_handler = TimedRotatingFileHandler(os.path.join(os.path.expanduser("~"), "QGrain", "logs", "qgrain.log"),
                                            when="D", backupCount=8, encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(format_str))
    status_bar_handler = StatusBarLogHandler(status_bar, level=logging.INFO)
    status_bar_handler.setFormatter(logging.Formatter(format_str))
    gui_handler = GUILogHandler(log_dialog, level=logging.DEBUG)
    gui_handler.setFormatter(logging.Formatter(format_str))
    logging.basicConfig(level=logging.DEBUG, format=format_str)
    logging.getLogger().addHandler(file_handler)
    logging.getLogger().addHandler(status_bar_handler)
    logging.getLogger().addHandler(gui_handler)
    mpl.set_loglevel("error")


def setup_app(language="en", theme="default"):
    assert language in ("en", "zh_CN")
    assert theme in ("default", "light", "dark")
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling)
    if not QtWidgets.QApplication.instance():
        app = QtWidgets.QApplication(sys.argv)
    else:
        app = QtWidgets.QApplication.instance()
    pixmap = QtGui.QPixmap(os.path.join(QGRAIN_ROOT_PATH, "assets", "icon.png"))
    create_necessary_folders()
    app.setWindowIcon(QtGui.QIcon(pixmap))
    app.setApplicationVersion(QGRAIN_VERSION)
    from qt_material import apply_stylesheet
    apply_stylesheet(app, theme=os.path.join(QGRAIN_ROOT_PATH, "assets", f"{theme}_theme.xml"),
                     invert_secondary=True, extra=EXTRA)
    setup_matplotlib()
    setup_language(app, f"{language}")
    return app



__all__ = ["CumulativeChart"]

from typing import *

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtGui, QtWidgets
from numpy import ndarray

from . import BaseChart
from ..models import Sample
from ..statistics import to_microns, to_cumulative


class CumulativeChart(BaseChart):
    def __init__(self, parent=None, figsize=(3, 2.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Cumulative Frequency Chart"))
        self._axes = self._figure.subplots()
        self.scale_menu = QtWidgets.QMenu(self.tr("Scale"))
        self.menu.insertMenu(self.edit_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions: List[QtGui.QAction] = []
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name)
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)
        self._last_samples = []

    @property
    def supported_scales(self) -> Sequence[Tuple[str, str]]:
        scales = (("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear")))
        return scales

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def transfer(self) -> Callable[[Union[int, float, ndarray]], Union[int, float, ndarray]]:
        if self.scale == "log-linear":
            return lambda classes_phi: to_microns(classes_phi)
        elif self.scale == "log":
            return lambda classes_phi: np.log(to_microns(classes_phi))
        elif self.scale == "phi":
            return lambda classes_phi: classes_phi
        elif self.scale == "linear":
            return lambda classes_phi: to_microns(classes_phi)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size (microns)"
        elif self.scale == "log":
            return "Ln(grain size in microns)"
        elif self.scale == "phi":
            return "Grain size (phi)"
        elif self.scale == "linear":
            return "Grain size (microns)"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def update_chart(self):
        self._figure.clear()
        self._axes = self._figure.subplots()
        self.show_samples(self._last_samples, append=False)

    def show_samples(self, samples: Sequence[Sample], append=False, title=""):
        if len(samples) == 0:
            return
        append = append and len(self._last_samples) != 0
        if not append:
            self._axes.clear()
            self._last_samples = []
            if self.xlog:
                self._axes.set_xscale("log")
            x = self.transfer(samples[0].classes_phi)
            self._axes.set_title(title)
            self._axes.set_xlabel(self.xlabel)
            self._axes.set_ylabel(self.ylabel)
            self._axes.set_xlim(x[0], x[-1])
            self._axes.set_ylim(0.0, 1.0)
        for i, sample in enumerate(samples):
            self._last_samples.append(sample)
            x = self.transfer(samples[0].classes_phi)
            cumulative_frequency = to_cumulative(sample.distribution)
            c = plt.get_cmap()(i % 10)
            self._axes.plot(x, cumulative_frequency, c=c, marker=".", mfc=c, mec=c, label=sample.name)
        self._figure.tight_layout()
        self._canvas.draw()

    def retranslate(self):
        self.setWindowTitle(self.tr("Cumulative Frequency Chart"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))
        self.scale_menu.setTitle(self.tr("Scale"))



from enum import Enum, unique


@unique
class DistributionType(Enum):
    Normal = 1
    SkewNormal = 2
    Weibull = 3
    GeneralWeibull = 4


@unique
class KernelType(Enum):
    Nonparametric = 0
    Normal = 1
    SkewNormal = 2
    Weibull = 3
    GeneralWeibull = 4


from .dataset import *
from .artificial_dataset import *
from .ssu_result import *
from .emma_result import *
from .udm_result import *
